import json
from drsai import AssistantAgent, HepAIChatCompletionClient
import asyncio
from typing import AsyncGenerator,Optional, List, Dict, Tuple, Sequence, Any, Awaitable, Callable, Union
from loguru import logger
# from datetime import datetime
from pydantic import BaseModel


from autogen_core import CancellationToken, FunctionCall
from autogen_core.model_context import (
    ChatCompletionContext,
)
from autogen_agentchat.base import Response, TaskResult
from autogen_agentchat.base import Handoff as HandoffBase
from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    # HandoffMessage,
    # MemoryQueryEvent,
    # ModelClientStreamingChunkEvent,
    # StructuredMessage,
    # StructuredMessageFactory,
    TextMessage,
    ThoughtEvent,
    ModelClientStreamingChunkEvent,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
    UserInputRequestedEvent,
)
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    CreateResult,
    RequestUsage,
    FunctionExecutionResult,
    FunctionExecutionResultMessage,
    LLMMessage,
    # ModelFamily,
    SystemMessage,
)
from autogen_core.tools import (
    BaseTool, 
    # FunctionTool, 
    # StaticWorkbench, 
    Workbench, 
    # ToolResult, 
    # TextResultContent, 
    # ToolSchema
    )

from ..magentic_one.guarded_action import ApprovalDeniedError
import aiohttp
from aiohttp import ClientConnectionError

class RemoteAgent(AssistantAgent):
    '''
    连接OpenAI格式的模型或者智能体后端
    '''
    def __init__(
            self, 
            name: str,
            model_client: HepAIChatCompletionClient|None = None,
            model_client_stream: bool = True,
            tools: List[BaseTool[Any, Any] | Callable[..., Any] | Callable[..., Awaitable[Any]]] | None = None,
            description: str = "An agent that provides assistance with ability to use tools.",
            system_message: (
                str | None
            ) = "You are a helpful AI assistant. Solve tasks using your tools. Reply with TERMINATE when the task has been completed.",
            memory_function: Callable | None = None,
            allow_reply_function: bool = False,
            reply_function: Callable | None = None,
            model_remote_configs: Dict[str, Any] = {},
            chat_id: str|None = None,
            run_info: Dict[str, Any] = {},
            **kwargs):
        
        super().__init__(
            name = name, 
            model_client = model_client,
            model_client_stream = model_client_stream,
            tools = tools,
            description = description,
            system_message = system_message,
            memory_function = memory_function,
            allow_reply_function = allow_reply_function,
            reply_function = reply_function,
            **kwargs)

        self.is_paused = False
        self._paused = asyncio.Event()
        self._current_streaming_response = None

        self._chat_id = chat_id
        run_info.update(kwargs)
        self._run_info = run_info 
        
        # initialize the async model client
        self.api_key = model_remote_configs.pop("api_key", "")
        self.url = model_remote_configs.pop("url", "")
        self.new_headers = {}
        self.new_headers["Authorization"] = f"Bearer {self.api_key}"
        self.new_headers["Content-Type"] = "application/json"
        self._session = None
        self._connection_timeout = 60

    async def lazy_init(self, **kwargs: Any) -> None:
        """Initialize the tools and models needed by the agent."""
        if self._session is None:
            self._session = aiohttp.ClientSession(
                # base_url=self.url, 如"http://localhost:42807/apiv2"，这里直接在on_messages_stream中使用完整的url
                headers=self.new_headers,
                timeout=aiohttp.ClientTimeout(total=self._connection_timeout)
            )

    async def close(self) -> None:
        """Clean up resources used by the agent.

        This method:
          ...
        """
        logger.info(f"Closing {self.name}...")

        # 中断当前流式响应
        if self._current_streaming_response:
            try:
                self._current_streaming_response.close()
                # logger.debug(f"Force-closed streaming response for {self.name}")
            except Exception as e:
                logger.warning(f"Error closing response: {str(e)}")

        # 关闭模型客户端
        if self._model_client:
            await self._model_client.close()
        
        # 关闭HTTP session
        if self._session and not self._session.closed:
            # await asyncio.sleep(0.5)
            await self._session.close()
        
        logger.info(f"Closed {self.name} successfully.")

    async def pause(self) -> None:
        """Pause the agent by setting the paused state."""
        logger.info(f"Paused {self.name}...")
        self.is_paused = True
        self._paused.set()

    async def resume(self) -> None:
        """Resume the agent by clearing the paused state."""
        self.is_paused = False
        self._paused.clear()

    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """Handle incoming messages and yield responses as a stream. Append the request to agents chat history."""

        # monitor the pause event
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content=f"The {self.name} is paused.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
            return

         # Set up the cancellation token for the this handler.
        code_execution_token = CancellationToken()

        # Cancel the task if the handler's cancellation token is set.
        cancellation_token.add_callback(lambda: code_execution_token.cancel())

        # Set up background task to monitor the pause event and cancel the task if paused.
        async def monitor_pause() -> None:
            await self._paused.wait()
            code_execution_token.cancel()
        monitor_pause_task = asyncio.create_task(monitor_pause())

        
        try:
        ##########Your costum code here##########
        # NOTE: Can only yield TextMessage or MultiModalMessage in MagenticAgent becasue the limit in src/magentic_ui/utils.py", line 160, in thread_to_context:
        # ```assert isinstance(m, TextMessage) or isinstance(m, MultiModalMessage)```

            # Gather all relevant state here
            agent_name = self.name
            model_context = self._model_context
            memory = self._memory
            system_messages = self._system_messages
            workbench = self._workbench
            handoff_tools = self._handoff_tools
            handoffs = self._handoffs
            model_client = self._model_client
            model_client_stream = self._model_client_stream
            reflect_on_tool_use = self._reflect_on_tool_use
            tool_call_summary_format = self._tool_call_summary_format
            output_content_type = self._output_content_type
            format_string = self._output_content_type_format

            # STEP 1: Add new user/handoff messages to the model context
            await self._add_messages_to_context(
                model_context=model_context,
                messages=messages,
            )

            # STEP 2: Update model context with any relevant memory
            inner_messages: List[BaseAgentEvent | BaseChatMessage] = []
            for event_msg in await self._update_model_context_with_memory(
                memory=memory,
                model_context=model_context,
                agent_name=agent_name,
            ):
                inner_messages.append(event_msg)
                yield event_msg

        
            # STEP 3: Run the first inference
            model_result = None
            all_messages = await model_context.get_messages()
            llm_messages: List[LLMMessage] = self._get_compatible_context(model_client=model_client, messages=system_messages + all_messages)
            oai_massages = await self.llm_messages2oai_messages(llm_messages)
            body = {
                "chat_id": self._chat_id, 
                "user": self._run_info,
                "model":agent_name, 
                "messages": oai_massages
                }
            
            # try:

            full_response = ""
            async with self._session.post(
                self.url,
                headers=self.new_headers,
                json=body
            ) as response:
                 
                self._current_streaming_response = response

                response.raise_for_status()
                
                buffer = b''
                async for chunk in response.content.iter_chunked(1024):
                    
                    # 检查取消和暂停状态
                    if code_execution_token.is_cancelled() or self.is_paused:
                        raise asyncio.CancelledError()

                    buffer += chunk
                    while b'\n' in buffer:
                        line_bytes, buffer = buffer.split(b'\n', 1)
                        line = line_bytes.decode().strip()
                        
                        if line.startswith("data: "):
                            try:
                                json_str = line[6:]  # 去掉"data: "前缀
                                oai_json = json.loads(json_str)
                                # 安全访问嵌套字段
                                if "choices" in oai_json and len(oai_json["choices"]) > 0:
                                    delta = oai_json["choices"][0].get("delta", {})
                                    textchunck = delta.get("content", "")
                                    if textchunck:
                                        yield ModelClientStreamingChunkEvent(content=textchunck, source=agent_name)
                                        full_response += textchunck
                            except (json.JSONDecodeError, KeyError) as parse_error:
                                logger.warning(f"Failed to parse SSE data: {str(parse_error)}")

            self._current_streaming_response = None
            model_result = CreateResult(
                content=full_response, 
                finish_reason="stop",
                usage = RequestUsage(prompt_tokens = 0, completion_tokens = len(full_response.split())),
                cached = False
                )
            
            # except aiohttp.ClientError as e:
            #     logger.debug(f"HTTP error: {str(e)}")
            #     raise RuntimeError(f"HTTP error: {str(e)}")

            # except asyncio.CancelledError:
            #     logger.debug("Streaming request cancelled")
            #     raise asyncio.CancelledError()

            assert model_result is not None, "No model result was produced."

            # --- NEW: If the model produced a hidden "thought," yield it as an event ---
            if model_result.thought:
                thought_event = ThoughtEvent(content=model_result.thought, source=agent_name)
                yield thought_event
                inner_messages.append(thought_event)

            # Add the assistant message to the model context (including thought if present)
            await model_context.add_message(
                AssistantMessage(
                    content=model_result.content,
                    source=agent_name,
                    thought=getattr(model_result, "thought", None),
                )
            )

            # STEP 4: Process the model output
            async for output_event in self._process_model_result(
                model_result=model_result,
                inner_messages=inner_messages,
                cancellation_token=cancellation_token,
                agent_name=agent_name,
                system_messages=system_messages,
                model_context=model_context,
                workbench=workbench,
                handoff_tools=handoff_tools,
                handoffs=handoffs,
                model_client=model_client,
                model_client_stream=model_client_stream,
                reflect_on_tool_use=reflect_on_tool_use,
                tool_call_summary_format=tool_call_summary_format,
                output_content_type=output_content_type,
                format_string=format_string,
            ):
                yield output_event

        ##########Your costum code above##########

        except ApprovalDeniedError:
            # If the user denies the approval, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The user did not approve the task.",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            if self._current_streaming_response:
                self._current_streaming_response.close()
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        
        except ClientConnectionError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )

        except Exception as e:
            logger.error(f"Error in {self.name}: {e}")
            # add to chat history
            await model_context.add_message(
                AssistantMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self.name
                )
            )
            yield Response(
                chat_message=TextMessage(
                    content="An error occurred while executing the task.",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        finally:

            self._current_streaming_response = None
            # Cancel the monitor task.
            try:
                monitor_pause_task.cancel()
                await monitor_pause_task
            except asyncio.CancelledError:
                pass
    
    @classmethod
    async def _process_model_result(
        cls,
        model_result: CreateResult,
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        cancellation_token: CancellationToken,
        agent_name: str,
        system_messages: List[SystemMessage],
        model_context: ChatCompletionContext,
        workbench: Workbench,
        handoff_tools: List[BaseTool[Any, Any]],
        handoffs: Dict[str, HandoffBase],
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        reflect_on_tool_use: bool,
        tool_call_summary_format: str,
        output_content_type: type[BaseModel] | None,
        format_string: str | None = None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        Handle final or partial responses from model_result, including tool calls, handoffs,
        and reflection if needed.

        NOTE: Can only yield TextMessage or MultiModalMessage in MagenticAgent becasue the limit in src/magentic_ui/utils.py", line 160, in thread_to_context:
        ```assert isinstance(m, TextMessage) or isinstance(m, MultiModalMessage)```

        """

        # If direct text response (string)
        if isinstance(model_result.content, str):
            # if output_content_type:
            #     content = output_content_type.model_validate_json(model_result.content)
            #     yield Response(
            #         chat_message=StructuredMessage[output_content_type](  # type: ignore[valid-type]
            #             content=content,
            #             source=agent_name,
            #             models_usage=model_result.usage,
            #             format_string=format_string,
            #         ),
            #         inner_messages=inner_messages,
            #     )
            # else:
            #     yield Response(
            #         chat_message=TextMessage(
            #             content=model_result.content,
            #             source=agent_name,
            #             models_usage=model_result.usage,
            #         ),
            #         inner_messages=inner_messages,
            #     )
            yield Response(
                chat_message=TextMessage(
                    content=model_result.content,
                    source=agent_name,
                    models_usage=model_result.usage,
                    metadata={"internal": "no"}, # detect if it is internal message or not
                ),
                inner_messages=inner_messages,
            )
            return

        # Otherwise, we have function calls
        assert isinstance(model_result.content, list) and all(
            isinstance(item, FunctionCall) for item in model_result.content
        )

        # cannot yield ToolCallRequestEvent in MagenticAgent
        # STEP 4A: Yield ToolCallRequestEvent
        tool_call_msg = ToolCallRequestEvent(
            content=model_result.content,
            source=agent_name,
            models_usage=model_result.usage,
        )

        # event_logger.debug(tool_call_msg)
        logger.info(f"{cls.name} handling tool call with params : {model_result.content}")
        inner_messages.append(tool_call_msg)
        yield tool_call_msg

        # STEP 4B: Execute tool calls
        executed_calls_and_results = await asyncio.gather(
            *[
                cls._execute_tool_call(
                    tool_call=call,
                    workbench=workbench,
                    handoff_tools=handoff_tools,
                    agent_name=agent_name,
                    cancellation_token=cancellation_token,
                )
                for call in model_result.content
            ]
        )

        # cannot yield ToolCallExecutionEvent in MagenticAgent
        exec_results = [result for _, result in executed_calls_and_results]

        # Yield ToolCallExecutionEvent
        tool_call_result_msg = ToolCallExecutionEvent(
            content=exec_results,
            source=agent_name,
        )
        # event_logger.debug(tool_call_result_msg)
        await model_context.add_message(FunctionExecutionResultMessage(content=exec_results))
        inner_messages.append(tool_call_result_msg)
        yield tool_call_result_msg

        # cannot yield HandoffMessage
        # # STEP 4C: Check for handoff
        # handoff_output = cls._check_and_handle_handoff(
        #     model_result=model_result,
        #     executed_calls_and_results=executed_calls_and_results,
        #     inner_messages=inner_messages,
        #     handoffs=handoffs,
        #     agent_name=agent_name,
        # )
        # if handoff_output:
        #     yield handoff_output
        #     return

        # STEP 4D: Reflect or summarize tool results
        if reflect_on_tool_use:
            async for reflection_response in cls._reflect_on_tool_use_flow(
                system_messages=system_messages,
                model_client=model_client,
                model_client_stream=model_client_stream,
                model_context=model_context,
                agent_name=agent_name,
                inner_messages=inner_messages,
                output_content_type=output_content_type,
            ):
                yield reflection_response
        else:
            yield cls._summarize_tool_use(
                executed_calls_and_results=executed_calls_and_results,
                inner_messages=inner_messages,
                handoffs=handoffs,
                tool_call_summary_format=tool_call_summary_format,
                agent_name=agent_name,
            )

    @classmethod
    async def _reflect_on_tool_use_flow(
        cls,
        system_messages: List[SystemMessage],
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        model_context: ChatCompletionContext,
        agent_name: str,
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[Response | ModelClientStreamingChunkEvent | ThoughtEvent, None]:
        """
        If reflect_on_tool_use=True, we do another inference based on tool results
        and yield the final text response (or streaming chunks).
        """
        all_messages = system_messages + await model_context.get_messages()
        llm_messages = cls._get_compatible_context(model_client=model_client, messages=all_messages)

        reflection_result: Optional[CreateResult] = None

        if model_client_stream:
            async for chunk in model_client.create_stream(
                llm_messages,
                json_output=output_content_type,
            ):
                if isinstance(chunk, CreateResult):
                    reflection_result = chunk
                elif isinstance(chunk, str):
                    yield ModelClientStreamingChunkEvent(content=chunk, source=agent_name)
                else:
                    raise RuntimeError(f"Invalid chunk type: {type(chunk)}")
        else:
            reflection_result = await model_client.create(llm_messages, json_output=output_content_type)

        if not reflection_result or not isinstance(reflection_result.content, str):
            raise RuntimeError("Reflect on tool use produced no valid text response.")

        # --- NEW: If the reflection produced a thought, yield it ---
        if reflection_result.thought:
            thought_event = ThoughtEvent(content=reflection_result.thought, source=agent_name)
            yield thought_event
            inner_messages.append(thought_event)

        # Add to context (including thought if present)
        await model_context.add_message(
            AssistantMessage(
                content=reflection_result.content,
                source=agent_name,
                thought=getattr(reflection_result, "thought", None),
            )
        )

        # if output_content_type:
        #     content = output_content_type.model_validate_json(reflection_result.content)
        #     yield Response(
        #         chat_message=StructuredMessage[output_content_type](  # type: ignore[valid-type]
        #             content=content,
        #             source=agent_name,
        #             models_usage=reflection_result.usage,
        #         ),
        #         inner_messages=inner_messages,
        #     )
        # else:
        #     yield Response(
        #         chat_message=TextMessage(
        #             content=reflection_result.content,
        #             source=agent_name,
        #             models_usage=reflection_result.usage,
        #         ),
        #         inner_messages=inner_messages,
        #     )

        yield Response(
            chat_message=TextMessage(
                content=reflection_result.content,
                source=agent_name,
                models_usage=reflection_result.usage,
                metadata={"internal": "no"}, # detect if it is internal message or not
            ),
            inner_messages=inner_messages,
        )

    @staticmethod
    def _summarize_tool_use(
        executed_calls_and_results: List[Tuple[FunctionCall, FunctionExecutionResult]],
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        handoffs: Dict[str, HandoffBase],
        tool_call_summary_format: str,
        agent_name: str,
    ) -> Response:
        """
        If reflect_on_tool_use=False, create a summary message of all tool calls.
        """
        # Filter out calls which were actually handoffs
        normal_tool_calls = [(call, result) for call, result in executed_calls_and_results if call.name not in handoffs]
        tool_call_summaries: List[str] = []
        for tool_call, tool_call_result in normal_tool_calls:
            # 对MCP的结果进行处理
            try:
                json_results = json.loads(tool_call_result.content)
                if isinstance(json_results, list):
                    json_result = json_results[0]
                    if isinstance(json_result, dict) and 'type' in json_result:
                        if json_result['type'] == 'text':
                            tool_call_result.content = json_result['text']
            except:
                pass
            tool_call_summaries.append(
                tool_call_summary_format.format(
                    tool_name=tool_call.name,
                    arguments=tool_call.arguments,
                    result=tool_call_result.content,
                )
            )
        tool_call_summary = "\n".join(tool_call_summaries)
        # return Response(
        #     chat_message=ToolCallSummaryMessage(
        #         content=tool_call_summary,
        #         source=agent_name,
        #     ),
        #     inner_messages=inner_messages,
        # )

        return Response(
            chat_message=TextMessage(
                content=tool_call_summary,
                source=agent_name,
                metadata={"internal": "no"}, # detect if it is internal message or not
            ),
            inner_messages=inner_messages,
        )
    
    async def run_stream(
        self,
        *,
        task: str | BaseChatMessage | Sequence[BaseChatMessage] | None = None,
        cancellation_token: CancellationToken | None = None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | TaskResult, None]:
        """Run the agent with the given task and return a stream of messages
        and the final task result as the last item in the stream."""
        if cancellation_token is None:
            cancellation_token = CancellationToken()
        input_messages: List[BaseChatMessage] = []
        output_messages: List[BaseAgentEvent | BaseChatMessage] = []
        if task is None:
            pass
        elif isinstance(task, str):
            text_msg = TextMessage(content=task, source="user", metadata={"internal": "yes"})
            input_messages.append(text_msg)
            output_messages.append(text_msg)
            yield text_msg
        elif isinstance(task, BaseChatMessage):
            input_messages.append(task)
            output_messages.append(task)
            task.metadata["internal"] = "yes"
            yield task
        else:
            if not task:
                raise ValueError("Task list cannot be empty.")
            for msg in task:
                if isinstance(msg, BaseChatMessage):
                    input_messages.append(msg)
                    output_messages.append(msg)
                    msg.metadata["internal"] = "yes"
                    yield msg
                else:
                    raise ValueError(f"Invalid message type in sequence: {type(msg)}")
        async for message in self.on_messages_stream(input_messages, cancellation_token):
            if isinstance(message, Response):
                yield message.chat_message
                output_messages.append(message.chat_message)
                yield TaskResult(messages=output_messages)
                
            else:
                yield message
                if isinstance(message, ModelClientStreamingChunkEvent):
                    # Skip the model client streaming chunk events.
                    continue
                output_messages.append(message)




import os
import json
import logging
import zipfile
import requests
from typing import List, Dict
from ...benchmark import Benchmark
from ...models import (
    BaseTask,
    BaseEvalResult,
    AllTaskTypes,
    AllCandidateTypes,
    AllEvalResultTypes,
)


class BearcubsBenchmark(Benchmark):
    """
    Loads the Bearcubs dataset from bear-cubs.github.io, stores it locally.
    Benchmark paper: https://arxiv.org/abs/2503.07919
    This benchmark has no evaluation since there are no ground truth answers.
    """

    DATASET_URL = "https://bear-cubs.github.io/BearCubs_20250310.json.zip"
    DATASET_FILE = "BearCubs_20250310.json"

    def __init__(self, name: str = "Bearcubs", data_dir: str | None = None):
        assert data_dir is not None, "data_dir must be provided for Bearcubs"
        super().__init__(name=name, data_dir=data_dir)
        self.eval_result_class = BaseEvalResult

    def download_dataset(self) -> None:
        """
        Download and unzip the dataset into self.data_dir
        """
        assert self.data_dir is not None
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir, exist_ok=True)

        zip_path = os.path.join(self.data_dir, "bearcubs.zip")
        json_path = os.path.join(self.data_dir, self.DATASET_FILE)

        # Skip if json already exists
        if os.path.exists(json_path):
            logging.info(f"[Bearcubs] Dataset already exists at '{json_path}'")
            return

        # Download zip file
        logging.info(f"[Bearcubs] Downloading dataset from '{self.DATASET_URL}'...")
        response = requests.get(self.DATASET_URL)
        response.raise_for_status()

        # Save zip file
        with open(zip_path, "wb") as f:
            f.write(response.content)

        # Extract zip file
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(self.data_dir)

        # Remove zip file
        os.remove(zip_path)
        logging.info("[Bearcubs] Dataset downloaded and extracted.")

    def load_dataset(self) -> None:
        """
        Read in the json file and store tasks in self.tasks
        """
        assert self.data_dir is not None
        json_path = os.path.join(self.data_dir, self.DATASET_FILE)

        if not os.path.exists(json_path):
            raise FileNotFoundError(
                f"Could not find {self.DATASET_FILE} in {self.data_dir}. "
                "Make sure you have downloaded the dataset."
            )

        # Load json file
        with open(json_path, "r", encoding="utf-8") as f:
            data = json.load(f)

        # Create tasks
        self.tasks: Dict[str, BaseTask] = {}
        for id_, item in data.items():
            task = BaseTask(
                id=id_,
                question=item["question"],
                ground_truth="",  # No ground truth answers
                set="test",  # Single test split
            )
            self.tasks[task.id] = task

        logging.info(f"[Bearcubs] Loaded {len(self.tasks)} examples.")

    def get_split_tasks(self, split: str) -> List[str]:
        """
        Returns task IDs for the specified split. Only 'test' split is available.
        """
        if split != "test":
            raise ValueError("Only 'test' split is available for Bearcubs")
        return [task_id for task_id, task in self.tasks.items() if task.set == split]

    def evaluator(
        self, task: AllTaskTypes, candidate: AllCandidateTypes
    ) -> AllEvalResultTypes:
        """
        Raises error since evaluation is not supported for this benchmark.
        """
        raise NotImplementedError(
            "Evaluation is not supported for Bearcubs benchmark as there are no ground truth answers"
        )




from autogen_core.models import ChatCompletionClient
from autogen_core import ComponentModel
from autogen_agentchat.base import ChatAgent, TaskResult, Team
from autogen_agentchat.agents import UserProxyAgent, BaseChatAgent, AssistantAgent
from autogen_agentchat.teams import BaseGroupChat

from .tools.playwright.browser import get_browser_resource_config
from .teams import GroupChat, RoundRobinGroupChat
from .teams.orchestrator.orchestrator_config import OrchestratorConfig
from .agents import WebSurfer, CoderAgent, USER_PROXY_DESCRIPTION, FileSurfer
from .magentic_ui_config import MagenticUIConfig, ModelClientConfigs
from .agents.web_surfer import WebSurferConfig
from .agents.users import DummyUserProxy, MetadataUserProxy
from .approval_guard import (
    ApprovalGuard,
    ApprovalGuardContext,
    ApprovalConfig,
    BaseApprovalGuard,
)
from ..remote_agent.drsai_remote_agent import RemoteAgent
# from ..magentic_one.agents.drsai_agents.drsai_agent import MagenticAgent
from drsai.modules.baseagent import DrSaiAgent
from ...agent_factory.load_agent import get_model_client

from ...ui_backend.input_func import InputFuncType, make_agentchat_input_func
from ...ui_backend.learning.memory_provider import MemoryControllerProvider
from ...ui_backend.utils import get_internal_urls
from ...ui_backend.types import RunPaths
from ...ui_backend.backend.datamodel.types import EnvironmentVariable
from ...ui_backend.backend.utils.utils import decompress_state
from ...agent_factory.remote_agent import StatusAgent

import json, os
import aiofiles
from pathlib import Path
from typing import (
    Any,
    cast,
    Dict,
    List,
    Mapping,
    Optional,
    Union,
    Callable,
)
from ..load_agent import a_load_agent_factory_from_config

import yaml
import importlib
from loguru import logger

async def get_task_team(
    magentic_ui_config: Optional[MagenticUIConfig] = None,
    input_func: Optional[InputFuncType] = None,
    *,
    paths: RunPaths,
) -> GroupChat | RoundRobinGroupChat:
    """
    Creates and returns a GroupChat team with specified configuration.

    Args:
        magentic_ui_config (MagenticUIConfig, optional): Magentic UI configuration for team. Default: None.
        paths (RunPaths): Paths for internal and external run directories.

    Returns:
        GroupChat | RoundRobinGroupChat: An instance of GroupChat or RoundRobinGroupChat with the specified agents and configuration.
    """
    if magentic_ui_config is None:
        magentic_ui_config = MagenticUIConfig()

    def get_model_client(
        model_client_config: Union[ComponentModel, Dict[str, Any], None],
        is_action_guard: bool = False,
    ) -> ChatCompletionClient:
        if model_client_config is None:
            return ChatCompletionClient.load_component(
                ModelClientConfigs.get_default_client_config()
                if not is_action_guard
                else ModelClientConfigs.get_default_action_guard_config()
            )
        return ChatCompletionClient.load_component(model_client_config)

    if not magentic_ui_config.inside_docker:
        assert (
            paths.external_run_dir == paths.internal_run_dir
        ), "External and internal run dirs must be the same in non-docker mode"

    model_client_orch = get_model_client(
        magentic_ui_config.model_client_configs.orchestrator
    )
    approval_guard: BaseApprovalGuard | None = None

    approval_policy = (
        magentic_ui_config.approval_policy
        if magentic_ui_config.approval_policy
        else "never"
    )

    websurfer_loop_team: bool = (
        magentic_ui_config.websurfer_loop if magentic_ui_config else False
    )

    model_client_coder = get_model_client(magentic_ui_config.model_client_configs.coder)
    model_client_file_surfer = get_model_client(
        magentic_ui_config.model_client_configs.file_surfer
    )
    browser_resource_config, _novnc_port, _playwright_port = (
        get_browser_resource_config(
            paths.external_run_dir,
            magentic_ui_config.novnc_port,
            magentic_ui_config.playwright_port,
            magentic_ui_config.inside_docker,
        )
    )

    orchestrator_config = OrchestratorConfig(
        cooperative_planning=magentic_ui_config.cooperative_planning,
        autonomous_execution=magentic_ui_config.autonomous_execution,
        allowed_websites=magentic_ui_config.allowed_websites,
        plan=magentic_ui_config.plan,
        model_context_token_limit=magentic_ui_config.model_context_token_limit,
        do_bing_search=magentic_ui_config.do_bing_search,
        retrieve_relevant_plans=magentic_ui_config.retrieve_relevant_plans,
        memory_controller_key=magentic_ui_config.memory_controller_key,
        allow_follow_up_input=magentic_ui_config.allow_follow_up_input,
        final_answer_prompt=magentic_ui_config.final_answer_prompt,
    )
    websurfer_model_client = magentic_ui_config.model_client_configs.web_surfer
    if websurfer_model_client is None:
        websurfer_model_client = ModelClientConfigs.get_default_client_config()
    websurfer_config = WebSurferConfig(
        name="web_surfer",
        model_client=websurfer_model_client,
        browser=browser_resource_config,
        single_tab_mode=False,
        max_actions_per_step=magentic_ui_config.max_actions_per_step,
        url_statuses={key: "allowed" for key in orchestrator_config.allowed_websites}
        if orchestrator_config.allowed_websites
        else None,
        url_block_list=get_internal_urls(magentic_ui_config.inside_docker, paths),
        multiple_tools_per_call=magentic_ui_config.multiple_tools_per_call,
        downloads_folder=str(paths.internal_run_dir),
        debug_dir=str(paths.internal_run_dir),
        animate_actions=True,
        start_page=None,
        use_action_guard=True,
        to_save_screenshots=False,
    )

    user_proxy: DummyUserProxy | MetadataUserProxy | UserProxyAgent

    if magentic_ui_config.user_proxy_type == "dummy":
        user_proxy = DummyUserProxy(name="user_proxy")
    elif magentic_ui_config.user_proxy_type == "metadata":
        assert (
            magentic_ui_config.task is not None
        ), "Task must be provided for metadata user proxy"
        assert (
            magentic_ui_config.hints is not None
        ), "Hints must be provided for metadata user proxy"
        assert (
            magentic_ui_config.answer is not None
        ), "Answer must be provided for metadata user proxy"
        user_proxy = MetadataUserProxy(
            name="user_proxy",
            description="Metadata User Proxy Agent",
            task=magentic_ui_config.task,
            helpful_task_hints=magentic_ui_config.hints,
            task_answer=magentic_ui_config.answer,
            model_client=model_client_orch,
        )
    else:
        user_proxy_input_func = make_agentchat_input_func(input_func)
        user_proxy = UserProxyAgent(
            description=USER_PROXY_DESCRIPTION,
            name="user_proxy",
            input_func=user_proxy_input_func,
        )

    if magentic_ui_config.user_proxy_type in ["dummy", "metadata"]:
        model_client_action_guard = get_model_client(
            magentic_ui_config.model_client_configs.action_guard,
            is_action_guard=True,
        )

        # Simple approval function that always returns yes
        def always_yes_input(prompt: str, input_type: str = "text_input") -> str:
            return "yes"

        approval_guard = ApprovalGuard(
            input_func=always_yes_input,
            default_approval=False,
            model_client=model_client_action_guard,
            config=ApprovalConfig(
                approval_policy=approval_policy,
            ),
        )
    elif input_func is not None:
        model_client_action_guard = get_model_client(
            magentic_ui_config.model_client_configs.action_guard
        )
        approval_guard = ApprovalGuard(
            input_func=input_func,
            default_approval=False,
            model_client=model_client_action_guard,
            config=ApprovalConfig(
                approval_policy=approval_policy,
            ),
        )
    with ApprovalGuardContext.populate_context(approval_guard):
        web_surfer = WebSurfer.from_config(websurfer_config)
    if websurfer_loop_team:
        # simplified team of only the web surfer
        team = RoundRobinGroupChat(
            participants=[web_surfer, user_proxy],
            max_turns=10000,
        )
        await team.lazy_init()
        return team

    coder_agent = CoderAgent(
        name="coder_agent",
        model_client=model_client_coder,
        work_dir=paths.internal_run_dir,
        bind_dir=paths.external_run_dir,
        model_context_token_limit=magentic_ui_config.model_context_token_limit,
        approval_guard=approval_guard,
    )

    file_surfer = FileSurfer(
        name="file_surfer",
        model_client=model_client_file_surfer,
        work_dir=paths.internal_run_dir,
        bind_dir=paths.external_run_dir,
        model_context_token_limit=magentic_ui_config.model_context_token_limit,
        approval_guard=approval_guard,
    )

    if (
        orchestrator_config.memory_controller_key is not None
        and orchestrator_config.retrieve_relevant_plans in ["reuse", "hint"]
    ):
        memory_provider = MemoryControllerProvider(
            internal_workspace_root=paths.internal_root_dir,
            external_workspace_root=paths.external_root_dir,
            inside_docker=magentic_ui_config.inside_docker,
        )
    else:
        memory_provider = None

    team = GroupChat(
        participants=[web_surfer, user_proxy, coder_agent, file_surfer],
        orchestrator_config=orchestrator_config,
        model_client=model_client_orch,
        memory_provider=memory_provider,
    )

    await team.lazy_init()
    return team

async def load_from_file(path: Union[str, Path]) -> Dict[str, Any]:
    """Load team configuration from JSON/YAML file"""
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Config file not found: {path}")

    async with aiofiles.open(path) as f:
        content = await f.read()
        if path.suffix == ".json":
            return json.loads(content)
        elif path.suffix in (".yml", ".yaml"):
            return yaml.safe_load(content)
        raise ValueError(f"Unsupported file format: {path.suffix}")

async def create_magentic_one_team(
    team_config: Union[str, Path, Dict[str, Any], ComponentModel],
    state: Optional[Mapping[str, Any] | str] = None,
    input_func: Optional[InputFuncType] = None,
    env_vars: Optional[List[EnvironmentVariable]] = None,
    settings_config: dict[str, Any] = {},
    *,
    paths: RunPaths,
    config: dict[str, Any],
    load_from_config: bool = False,
    inside_docker: bool = True,
) -> tuple[Team, int, int]:
    """Create team instance from config"""

    _, novnc_port, playwright_port = get_browser_resource_config(
        paths.external_run_dir, -1, -1, inside_docker
    )
    try:
        if not load_from_config:
            # Load model configurations from settings if provided
            model_configs: Dict[str, Any] = settings_config.get("model_configs")
            # Use model configs from settings if available, otherwise fall back to config
            orchestrator_config = model_configs.get("orchestrator_client", config.get("orchestrator_client", None))
            model_client_configs = ModelClientConfigs(
                orchestrator=orchestrator_config,
                web_surfer=model_configs.get(
                    "web_surfer_client",
                    config.get("web_surfer_client", None),
                ),
                coder=model_configs.get(
                    "coder_client", config.get("coder_client", None)
                ),
                file_surfer=model_configs.get(
                    "file_surfer_client",
                    config.get("file_surfer_client", None),
                ),
                action_guard=model_configs.get(
                    "action_guard_client",
                    config.get("action_guard_client", None),
                ),
            )

            magentic_ui_config = MagenticUIConfig(
                **(settings_config or {}),
                model_client_configs=model_client_configs,
                playwright_port=playwright_port,
                novnc_port=novnc_port,
                inside_docker=inside_docker,
            )

            team = cast(
                Team,
                await get_task_team(
                    magentic_ui_config=magentic_ui_config,
                    input_func=input_func,
                    paths=paths,
                ),
            )
            if hasattr(team, "_participants"):
                for agent in cast(list[ChatAgent], team._participants):  # type: ignore
                    if isinstance(agent, WebSurfer):
                        novnc_port = agent.novnc_port
                        playwright_port = agent.playwright_port

            if state:
                if isinstance(state, str):
                    try:
                        # Try to decompress if it's compressed
                        state_dict = decompress_state(state)
                        await team.load_state(state_dict)
                    except Exception:
                        # If decompression fails, assume it's a regular JSON string
                        state_dict = json.loads(state)
                        await team.load_state(state_dict)
                else:
                    await team.load_state(state)

            return team, novnc_port, playwright_port

        if isinstance(team_config, (str, Path)):
            config = await load_from_file(team_config)
        elif isinstance(team_config, dict):
            config = team_config
        else:
            config = team_config.model_dump()

        # Load env vars into environment if provided
        if env_vars:
            logger.info("Loading environment variables")
            for var in env_vars:
                os.environ[var.name] = var.value

        team = cast(Team, GroupChat.load_component(config))

        if hasattr(team, "_participants"):
            for agent in cast(list[ChatAgent], self.team._participants):  # type: ignore
                if hasattr(agent, "input_func"):
                    agent.input_func = input_func  # type: ignore
                if isinstance(agent, WebSurfer):
                    novnc_port = agent.novnc_port or -1
                    playwright_port = agent.playwright_port or -1
        return team, novnc_port, playwright_port
    except Exception as e:
        logger.error(f"Error creating team: {e}")
        raise

async def create_magentic_round_team(
    team_config: Union[str, Path, Dict[str, Any], ComponentModel],
    state: Optional[Mapping[str, Any] | str] = None,
    input_func: Optional[InputFuncType] = None,
    env_vars: Optional[List[EnvironmentVariable]] = None,
    settings_config: dict[str, Any] = {},
    *,
    paths: RunPaths,
    config: dict[str, Any],
    load_from_config: bool = False,
    inside_docker: bool = True,
    run_info: dict[str, Any] = {},
    agent_mode_config: dict[str, Any] = {},
    files: List[Dict[str, Any]] | None = None,
) -> tuple[Team, int, int]:
    
    model_configs: Dict[str, Any] = settings_config.get("model_configs")
    model_config = model_configs.get("model_config", {})
    api_key = model_config.get("config", {}).get("api_key")
    if api_key is None:
        raise ValueError("API key not provided in model_configs")

    # 目前可选的mode有，besiii，drsai，custom，magentic-one，remote
    agent_mode: str = agent_mode_config.get("mode", "besiii") 
    agent_config: Dict[str, Any] = agent_mode_config.get("config", {})
    
    # 配置chat_id
    user_id = run_info.pop("user_id")
    session_id = run_info.get("session_id")
    chat_id = str(user_id)+str(session_id)
    # 配置用户信息
    run_info["name"] = user_id
    run_info["email"] = user_id
    
    if agent_mode == "besiii":
        # raise NotImplementedError("BesIII mode not implemented yet")
        agent = StatusAgent(
            name='besiii',
            model_client=get_model_client(model_client_config = model_config),
            chat_id=chat_id,
            run_info=run_info,
            model_remote_configs={
                # "url": "http://0.0.0.0:42806/apiv2/",
                "model": "drsai/besiii",
                "api_key": api_key
            },
            files = files,
        )
        # agent = RemoteAgent(
        #     name='besiii',
        #     chat_id=chat_id,
        #     run_info=run_info,
        #     model_remote_configs={
        #         "url": "http://202.122.37.163:42887/apiv2/chat/completions",
        #         "model_name": "hepai/drsai",
        #         "api_key": api_key
        #     },
        # )
    
    elif agent_mode == "drsai":
        # raise NotImplementedError("DrSai mode not implemented yet")
        agent = DrSaiAgent(
            name='drsai',
            system_message="A helpfull assistant.",
            description="A helpfull assistant.",
            model_client=get_model_client(model_client_config = model_config)
        )
        
    elif agent_mode == "custom":
        agent_factory: Callable[[], Union[AssistantAgent, BaseGroupChat]] = await a_load_agent_factory_from_config(agent_config, mode = "ui")
        agent: AssistantAgent|BaseGroupChat = await agent_factory()

    elif agent_mode == "remote":
        agent_config["api_key"] = agent_config.get("api_key", api_key)
        agent = StatusAgent(
            name="RemoteAgent",
            model_client=get_model_client(model_client_config = model_config),
            model_remote_configs = agent_config,
            chat_id=chat_id,
            run_info=run_info
            )
    elif agent_mode == "pip_install":
        module = importlib.import_module(agent_config["provider"])
        agent_factory: Callable[[], Union[AssistantAgent, BaseGroupChat]] = await module.a_load_agent_factory_from_installed(
            team_config = team_config,
            state = state,
            input_func = input_func,
            env_vars = env_vars,
            settings_config = settings_config,
            paths = paths,
            config = config,
            load_from_config = load_from_config,
            inside_docker = inside_docker,
            run_info = run_info,
        )
        agent: AssistantAgent|BaseGroupChat = await agent_factory()

    else:
        agent_factory: Callable[[], Union[AssistantAgent, BaseGroupChat]] = await a_load_agent_factory_from_config(model_configs, mode = "ui")
        agent: AssistantAgent|BaseGroupChat = await agent_factory()

    if isinstance(agent, AssistantAgent):
        user_proxy_input_func = make_agentchat_input_func(input_func)
        user_proxy = UserProxyAgent(
            description=USER_PROXY_DESCRIPTION,
            name="user_proxy",
            input_func=user_proxy_input_func,
        )
        team = RoundRobinGroupChat(
            participants=[agent, user_proxy,],
        )
        
    else:
        logger.error(f"Only supports AssistantAgent")
        raise NotImplementedError("GroupChat mode not implemented yet")
    
    await team.lazy_init()

    if state:
        if isinstance(state, str):
            try:
                # Try to decompress if it's compressed
                state_dict = decompress_state(state)
                await team.load_state(state_dict)
            except Exception:
                # If decompression fails, assume it's a regular JSON string
                state_dict = json.loads(state)
                await team.load_state(state_dict)
        else:
            await team.load_state(state)

    return team, -1, -1


async def create_magentic_general_team(
    team_config: Union[str, Path, Dict[str, Any], ComponentModel],
    state: Optional[Mapping[str, Any] | str] = None,
    input_func: Optional[InputFuncType] = None,
    env_vars: Optional[List[EnvironmentVariable]] = None,
    settings_config: dict[str, Any] = {},
    *,
    paths: RunPaths,
    config: dict[str, Any],
    load_from_config: bool = False,
    inside_docker: bool = True,
) -> tuple[Team, int, int]:
    '''通过本地PIP安装的智能体/多智能体系统进行通用加载'''
    ...


from typing import Dict, List, Literal
import tldextract
from urllib.parse import urlparse

URL_ALLOWED: Literal["allowed"] = "allowed"
URL_REJECTED: Literal["rejected"] = "rejected"

UrlStatus = Literal["allowed", "rejected"]


class UrlStatusManager:
    """
    A class to manage URL access control through allow/reject lists and explicit blocking.

    The URL matching is hierarchical and follows these rules:
    1. Block list is checked first - if URL matches any blocked pattern, access is denied
    2. For remaining URLs, if no status list is defined (None), all URLs are allowed
    3. Otherwise, URL must explicitly match an allowed pattern and not match any rejected patterns

    Note:
        Overlapping URLs with different statuses will result in undefined behavior.
        Example: { "example.com": "allowed", "example.com/foo": "rejected" }
    """

    url_statuses: Dict[str, UrlStatus] | None
    # TODO: There's a lot of logic around url_statuses being None. Use a separate list to check if a url is explicitly blocked
    url_block_list: List[str] | None

    def __init__(
        self,
        url_statuses: Dict[str, UrlStatus] | None = None,
        url_block_list: List[str] | None = None,
    ) -> None:
        """
        Args:
            url_statuses (Dict[str, UrlStatus], optional): initial url status settings. All urls are valid if None. Default: None.
            url_block_list (List[str], optional): initial url block list. Default: None.
        """
        self.url_statuses = None
        # a little bit of a hack to make sure there are no trailing slashes, since they mess with the comparison later on
        if url_statuses is not None:
            self.url_statuses = {
                key.rstrip("/"): value for key, value in url_statuses.items()
            }

        self.url_block_list = url_block_list

    def set_url_status(self, url: str, status: UrlStatus) -> None:
        """
        Adds a website to the manager. No-op if initialization parameter was None

        Args:
            url (str): The website to add.
            status (UrlStatus): The status of the url. Can be either "allowed" or "rejected".
        """
        if self.url_statuses is not None:
            url = url.strip()
            # Trailing slash messes up the comparison later on
            url = url.rstrip("/")
            self.url_statuses[url] = status

    def _is_url_match(self, registered_url: str, proposed_url: str) -> bool:
        """
        Checks if a proposed URL matches a registered URL pattern.

        Args:
            registered_url (str): The registered URL pattern to match against.
            proposed_url (str): The proposed URL to check.

        Returns:
            bool: True if the proposed URL matches the registered URL pattern, False otherwise.
        """
        # If no scheme is provided, assume http
        if not urlparse(registered_url).scheme:
            registered_url = "http://" + registered_url
        if not urlparse(proposed_url).scheme:
            proposed_url = "http://" + proposed_url

        parsed_registered_url = urlparse(registered_url)
        parsed_proposed_url = urlparse(proposed_url)
        extracted_registered_url = tldextract.extract(registered_url)
        extracted_proposed_url = tldextract.extract(proposed_url)

        # if both urls have a scheme, check if they are the same (http and https are treated as the same)
        http_equivalent_schemes = ["http", "https"]
        if (
            parsed_registered_url.scheme in http_equivalent_schemes
            and parsed_proposed_url.scheme in http_equivalent_schemes
        ):
            pass
        elif parsed_registered_url.scheme != parsed_proposed_url.scheme:
            return False

        # Check each component of the URL
        # TODO: what to do about params, query, and fragment components?
        if extracted_registered_url.subdomain:
            if extracted_registered_url.subdomain != extracted_proposed_url.subdomain:
                return False
        if extracted_registered_url.domain != extracted_proposed_url.domain:
            return False
        if (
            extracted_registered_url.suffix
            and extracted_proposed_url.suffix != extracted_registered_url.suffix
        ):
            return False
        if parsed_registered_url.path:
            if not parsed_proposed_url.path.startswith(parsed_registered_url.path):
                return False

        return True

    def is_url_blocked(self, url: str) -> bool:
        """
        Checks if a url is explicitly blocked.

        Args:
            url (str): The website to check.

        Returns:
            bool: True if the url is blocked, False otherwise.
        """
        if self.url_block_list is None:
            return False
        if any(self._is_url_match(site, url) for site in self.url_block_list):
            return True
        return False

    def is_url_rejected(self, url: str) -> bool:
        """
        Checks if the user explicitly rejected approval of a url. This function does NOT check against the allowed list.

        Args:
            url (str): The website to check.

        Returns:
            bool: True if the url was rejected by the user, False otherwise.
        """
        if self.is_url_blocked(url):
            return True
        if self.url_statuses is None:
            return False
        if any(
            self._is_url_match(site, url) and self.url_statuses[site] == URL_REJECTED
            for site in self.url_statuses
        ):
            return True
        return False

    def is_url_allowed(self, url: str) -> bool:
        """
        Checks if a website is allowed. This function does NOT check against the rejected list

        Args:
            url (str): The website to check.

        Returns:
            bool: True if the url is allowed, False otherwise.
        """
        if self.is_url_blocked(url):
            return False
        if self.url_statuses is None:
            return True
        if not any(
            self._is_url_match(site, url) and self.url_statuses[site] == URL_ALLOWED
            for site in self.url_statuses
        ):
            return False
        return True

    def get_allowed_sites(self) -> List[str] | None:
        """
        Returns a list of all allowed sites.

        Returns:
            List[str] | None: A list of all allowed sites. None if all sites are allowed.
        """
        if not self.url_statuses:
            return None
        return [
            site for site in self.url_statuses if self.url_statuses[site] == URL_ALLOWED
        ]

    def get_rejected_sites(self) -> List[str] | None:
        """
        Returns a list of all rejected sites.
        Returns:
            List[str] | None: A list of all rejected sites. None if all sites are allowed.
        """
        if not self.url_statuses:
            return None
        return [
            site
            for site in self.url_statuses
            if self.url_statuses[site] == URL_REJECTED
        ]

    def get_blocked_sites(self) -> List[str] | None:
        """
        Returns a list of all blocked sites.
        Returns:
            List[str] | None: A list of all blocked sites. None if there are no blocked sites.
        """
        return self.url_block_list



from typing import Dict, Any, List, Optional, Union
from pydantic import BaseModel, Field


class BaseTask(BaseModel):
    """Base task model that all benchmark-specific tasks inherit from"""

    id: str
    question: str
    ground_truth: str = ""
    set: str
    metadata: Dict[str, Any] = Field(default_factory=dict)
    url_path: str = ""
    file_name: str = ""
    file_dir: str = ""


class BaseCandidate(BaseModel):
    """Base candidate model that all benchmark-specific candidates inherit from"""

    answer: str


class BaseEvalResult(BaseModel):
    """Base evaluation result model that all benchmark-specific results inherit from"""

    score: Union[float, Dict[str, float]]
    metadata: Dict[str, Any] = Field(default_factory=dict)


# AssistantBench specific models
class AssistantBenchTask(BaseTask):
    difficulty: str = ""
    explanation: str = ""
    gold_url: str = ""


class AssistantBenchCandidate(BaseCandidate):
    pass  # Uses base answer field only


class AssistantBenchEvalResult(BaseEvalResult):
    pass  # Uses base score field only


# GAIA specific models
class GaiaTask(BaseTask):
    difficulty: str = ""
    file_name: str = ""


class GaiaCandidate(BaseCandidate):
    pass  # Uses base answer field only


class GaiaEvalResult(BaseEvalResult):
    pass  # Uses base score field only


# WebVoyager specific models
class WebVoyagerTask(BaseTask):
    web_name: str = ""
    web: str = ""
    answer_type: Optional[str] = None


class WebVoyagerCandidate(BaseCandidate):
    screenshots: List[str] = Field(default_factory=list)


class WebVoyagerEvalResult(BaseEvalResult):
    reasoning: str = ""
    pass  # Uses base score field only


# Custom benchmark specific models
class CustomTask(BaseTask):
    url: str = ""
    target_final_url: str = ""
    intermediate_url_list: List[str] = Field(default_factory=list)


class CustomCandidate(BaseCandidate):
    target_final_url: str = ""
    intermediate_url_list: List[str] = Field(default_factory=list)


class CustomEvalResult(BaseEvalResult):
    pass


# Union types for all tasks, candidates, and eval results
AllTaskTypes = Union[BaseTask, AssistantBenchTask, GaiaTask, WebVoyagerTask, CustomTask]

AllCandidateTypes = Union[
    BaseCandidate,
    AssistantBenchCandidate,
    GaiaCandidate,
    WebVoyagerCandidate,
    CustomCandidate,
]

AllEvalResultTypes = Union[
    BaseEvalResult,
    AssistantBenchEvalResult,
    GaiaEvalResult,
    WebVoyagerEvalResult,
    CustomEvalResult,
]



import docker
import os
import sys
from docker.errors import DockerException, ImageNotFound

_PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))
VNC_BROWSER_IMAGE = "magentic-ui-vnc-browser"
PYTHON_IMAGE = "magentic-ui-python-env"

VNC_BROWSER_BUILD_CONTEXT = "magentic-ui-browser-docker"
PYTHON_BUILD_CONTEXT = "magentic-ui-python-env"


def check_docker_running() -> bool:
    try:
        client = docker.from_env()
        client.ping()  # type: ignore
        return True
    except (DockerException, ConnectionError):
        return False


def build_image(
    image_name: str, build_context: str, client: docker.DockerClient
) -> None:
    for segment in client.api.build(
        path=build_context,
        dockerfile="Dockerfile",
        rm=True,
        tag=image_name,
        decode=True,
    ):
        if "stream" in segment:
            lines = segment["stream"].splitlines()
            for line in lines:
                if line:
                    sys.stdout.write(line + "\n")
                    sys.stdout.flush()


def check_docker_image(image_name: str, client: docker.DockerClient) -> bool:
    try:
        client.images.get(image_name)
        return True
    except ImageNotFound:
        return False


def build_browser_image(client: docker.DockerClient | None = None) -> None:
    if client is None:
        client = docker.from_env()
    client = docker.from_env()
    build_image(
        VNC_BROWSER_IMAGE + ":latest",
        os.path.join(_PACKAGE_DIR, "docker", VNC_BROWSER_BUILD_CONTEXT),
        client,
    )


def build_python_image(client: docker.DockerClient | None = None) -> None:
    if client is None:
        client = docker.from_env()
    client = docker.from_env()
    build_image(
        PYTHON_IMAGE + ":latest",
        os.path.join(_PACKAGE_DIR, "docker", PYTHON_BUILD_CONTEXT),
        client,
    )


def check_browser_image(client: docker.DockerClient | None = None) -> bool:
    if client is None:
        client = docker.from_env()
    client = docker.from_env()
    return check_docker_image(VNC_BROWSER_IMAGE, client)


def check_python_image(client: docker.DockerClient | None = None) -> bool:
    if client is None:
        client = docker.from_env()
    client = docker.from_env()
    return check_docker_image(PYTHON_IMAGE, client)



import asyncio
import io
import json
import logging
import os
import re
import time
from datetime import datetime
import tldextract
from typing import (
    Any,
    AsyncGenerator,
    BinaryIO,
    Dict,
    List,
    Optional,
    Sequence,
    cast,
    Mapping,
    Union,
    Tuple,
    Literal,
)
from typing_extensions import Self
from loguru import logger
from urllib.parse import quote_plus
from pydantic import Field
import PIL.Image
import tiktoken
from autogen_agentchat.agents import BaseChatAgent
from autogen_agentchat.base import Response
from autogen_agentchat.messages import (
    BaseChatMessage,
    MultiModalMessage,
    TextMessage,
)
from autogen_agentchat.utils import content_to_str, remove_images
from autogen_agentchat.state import BaseState
from autogen_core import EVENT_LOGGER_NAME, CancellationToken, FunctionCall
from autogen_core import Image as AGImage
from autogen_core.tools import ToolSchema

from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    LLMMessage,
    RequestUsage,
    SystemMessage,
    UserMessage,
)
from autogen_core.model_context import TokenLimitedChatCompletionContext
from pydantic import BaseModel
from autogen_core import Component, ComponentModel

from playwright.async_api import (
    BrowserContext,
    Download,
    Page,
)

from ...tools.playwright.browser import PlaywrightBrowser, VncDockerPlaywrightBrowser

from ...approval_guard import (
    ApprovalGuardContext,
    BaseApprovalGuard,
    MaybeRequiresApproval,
)

from ._events import WebSurferEvent
from ._prompts import (
    WEB_SURFER_OCR_PROMPT,
    WEB_SURFER_QA_PROMPT,
    WEB_SURFER_QA_SYSTEM_MESSAGE,
    WEB_SURFER_TOOL_PROMPT,
    WEB_SURFER_SYSTEM_MESSAGE,
    WEB_SURFER_NO_TOOLS_PROMPT,
)
from ._set_of_mark import add_set_of_mark
from ._tool_definitions import (
    TOOL_CLICK,
    TOOL_HISTORY_BACK,
    TOOL_HOVER,
    TOOL_PAGE_DOWN,
    TOOL_PAGE_UP,
    TOOL_READ_PAGE_AND_ANSWER,
    TOOL_SLEEP,
    TOOL_TYPE,
    TOOL_VISIT_URL,
    TOOL_WEB_SEARCH,
    TOOL_STOP_ACTION,
    TOOL_SELECT_OPTION,
    TOOL_CREATE_TAB,
    TOOL_SWITCH_TAB,
    TOOL_CLOSE_TAB,
    TOOL_KEYPRESS,
    TOOL_REFRESH_PAGE,
    # TOOL_UPLOAD_FILE,
    # TOOL_CLICK_FULL,
)

from ...tools.tool_metadata import get_tool_metadata, ToolMetadata
from ...tools.playwright.types import InteractiveRegion
from ...tools.playwright.playwright_controller import PlaywrightController
from ...tools.playwright.playwright_state import (
    BrowserState,
    save_browser_state,
    load_browser_state,
)
from ...tools.url_status_manager import (
    UrlStatusManager,
    UrlStatus,
    URL_ALLOWED,
    URL_REJECTED,
)


# New configuration class for WebSurfer
class WebSurferConfig(BaseModel):
    # TODO: Add playwright and context
    name: str
    model_client: ComponentModel | Dict[str, Any]
    browser: ComponentModel | Dict[str, Any]
    model_context_token_limit: int | None = None
    downloads_folder: str | None = None
    description: str | None = None
    debug_dir: str | None = None
    start_page: str | None = "about:blank"
    animate_actions: bool = False
    to_save_screenshots: bool = False
    max_actions_per_step: int = 5
    to_resize_viewport: bool = True
    url_statuses: Dict[str, UrlStatus] | None = None
    url_block_list: List[str] | None = None
    single_tab_mode: bool = False
    json_model_output: bool = False
    multiple_tools_per_call: bool = False
    viewport_height: int = 1440
    viewport_width: int = 1440
    use_action_guard: bool = False


class WebSurferState(BaseState):
    """
    State class for saving and loading the WebSurfer's state.

    Attributes:
        chat_history (List[LLMMessage]): List of chat messages exchanged with the model.
        type (str, optional): The type of the state. Default: `WebSurferState`.
        browser_state (BrowserState, optional): The state of the browser, including tabs and active page.
    """

    chat_history: List[LLMMessage]
    type: str = Field(default="WebSurferState")
    browser_state: BrowserState | None = None


class WebSurfer(BaseChatAgent, Component[WebSurferConfig]):
    """An agent that can browse the web and interact with web pages using a browser.

    Works with any model that supports tool calling or JSON outputs. Works best with multimodal models.

    The WebSurfer uses Playwright to control a browser and navigate web pages. It can:
    - Visit URLs and perform web searches
    - Click links and buttons
    - Fill in forms and type text
    - Scroll and hover over elements
    - Read and extract content from pages
    - Take screenshots and handle downloads
    - Manage multiple browser tabs

    The agent uses an LLM to decide what actions to take based on the current page state
    and user instructions.

    The browser interactions are controlled through a set of tools that the LLM can invoke.
    The agent captures screenshots and page content to give the LLM context about the
    current state of the page.

    Args:
        name (str): The name of the agent.
        model_client (ChatCompletionClient): The model client used by the agent.
        browser (PlaywrightBrowser): The browser resource to use.
        downloads_folder (str, optional): The folder where downloads are saved. Default: None.
        description (str, optional): The description of the agent. Default: WebSurfer.DEFAULT_DESCRIPTION.
        debug_dir (str, optional): The directory where debug information is saved. Default: None.
        start_page (str, optional): The start page for the browser. Defaults: WebSurfer.DEFAULT_START_PAGE.
        animate_actions (bool, optional): Whether to animate actions. Default: False.
        to_save_screenshots (bool, optional): Whether to save screenshots. Default: False.
        max_actions_per_step (int, optional): The maximum number of actions per step. Default: 5.
        to_resize_viewport (bool, optional): Whether to resize the viewport. Default: True.
        url_statuses (Dict[str, Literal["allowed", "rejected"]], optional): The set of allowed and rejected websites. Default: None.
        single_tab_mode (bool, optional): Whether to use single tab mode. Default: False.
        url_block_list (List[str], optional): A list of URLs to block. Default: None.
        json_model_output (bool, optional): Whether to use JSON output for model_client instead of tool calls. Default: False.
        multiple_tools_per_call (bool, optional): Whether to allow execution of multiple tool calls sequentially per model call. Default: False.
        viewport_height (int, optional): The height of the viewport. Default: 1440.
        viewport_width (int, optional): The width of the viewport. Default: 1440.
    """

    component_type = "agent"
    component_config_schema = WebSurferConfig
    component_provider_override = "magentic_ui.agents.web_surfer.WebSurfer"

    DEFAULT_DESCRIPTION = """
    The websurfer has access to a web browser that it can control.
    It understands images and can use them to help it complete the task.

    In a single step when you ask the agent to do something, it will perform multiple actions sequentially until it decides to stop.
    The actions it can perform are:
    - visiting a web page url
    - performing a web search using Bing
    - Interact with a web page: clicking a button, hovering over a button, typing in a field, scrolling the page, select an option in a dropdown
    - Downloading a file from the web page and upload file from the local file system
    - Pressing a key on the keyboard to interact with the web page
    - answer a question based on the content of the entire page beyond just the vieport
    - wait on a page to load
    - interact with tabs on the page: closing a tab, switching to a tab, creating a new tab
    - refresh the page

    It can do multiple actions in a single step, but it will stop after it has performed the maximum number of actions or once it decides to stop.
    As mentioned, it can perform multiple actions per command given to it, for instance if asked to fill a form, it can input the first name, then the last name, then the email, and then submit, and then stop.
    """
    DEFAULT_START_PAGE = "about:blank"

    # Size of the image we send to the MLM
    # Current values represent a 0.85 scaling to fit within the GPT-4v short-edge constraints (768px)
    MLM_HEIGHT = 765
    MLM_WIDTH = 1224

    SCREENSHOT_TOKENS = 1105

    def __init__(
        self,
        name: str,
        model_client: ChatCompletionClient,
        browser: PlaywrightBrowser,
        model_context_token_limit: int | None = None,
        downloads_folder: str | None = None,
        description: str = DEFAULT_DESCRIPTION,
        debug_dir: str | None = None,
        start_page: str = DEFAULT_START_PAGE,
        animate_actions: bool = False,
        to_save_screenshots: bool = False,
        max_actions_per_step: int = 5,
        to_resize_viewport: bool = True,
        url_statuses: Optional[Dict[str, UrlStatus]] = None,
        url_block_list: Optional[List[str]] = None,
        single_tab_mode: bool = False,
        json_model_output: bool = False,
        multiple_tools_per_call: bool = False,
        viewport_height: int = 1440,
        viewport_width: int = 1440,
        use_action_guard: bool = False,
    ) -> None:
        """
        Initialize the WebSurfer.
        """
        super().__init__(name, description)
        if debug_dir is None and to_save_screenshots:
            raise ValueError(
                "Cannot save screenshots without a debug directory. Set it using the 'debug_dir' parameter. The debug directory is created if it does not exist."
            )
        self._model_client = model_client
        self._model_context = TokenLimitedChatCompletionContext(
            model_client=self._model_client, token_limit=model_context_token_limit
        )
        self.start_page = start_page
        self.downloads_folder = downloads_folder
        self.debug_dir = debug_dir
        self.to_save_screenshots = to_save_screenshots
        self.to_resize_viewport = to_resize_viewport
        self.animate_actions = animate_actions
        self.max_actions_per_step = max_actions_per_step
        self.single_tab_mode = single_tab_mode
        self.json_model_output = json_model_output
        # If the model does not support function calling, we will use JSON output
        # override the json_model_output flag
        if not self._model_client.model_info["function_calling"]:
            self.json_model_output = True
        self.multiple_tools_per_call = multiple_tools_per_call
        self.viewport_height = viewport_height
        self.viewport_width = viewport_width
        self.use_action_guard = use_action_guard
        self._browser = browser
        # Call init to set these in case not set
        self._context: BrowserContext | None = None
        self._url_status_manager: UrlStatusManager = UrlStatusManager(
            url_statuses=url_statuses, url_block_list=url_block_list
        )
        self._page: Page | None = None
        self._last_download: Download | None = None
        self._prior_metadata_hash: str | None = None
        self.logger = logging.getLogger(EVENT_LOGGER_NAME + f".{self.name}.WebSurfer")
        self._chat_history: List[LLMMessage] = []
        self._last_outside_message: str = ""
        self._last_rejected_url: str | None = None

        # TODO: These are a little bit of a hack so we can get the ports out of the browser object
        self.novnc_port = -1
        self.playwright_port = -1
        if isinstance(self._browser, VncDockerPlaywrightBrowser):
            self.novnc_port = self._browser.novnc_port
            self.playwright_port = self._browser.playwright_port

        # explicitly allow about:blank
        if not self._url_status_manager.is_url_allowed("about:blank"):
            self._url_status_manager.set_url_status("about:blank", URL_ALLOWED)
        # explicitly allow chrome-error://chromewebdata
        if not self._url_status_manager.is_url_allowed("chrome-error://chromewebdata"):
            self._url_status_manager.set_url_status(
                "chrome-error://chromewebdata", URL_ALLOWED
            )

        if not self._url_status_manager.is_url_allowed(self.start_page):
            self.start_page = "about:blank"
            self.logger.warning(
                f"Default start page '{self.DEFAULT_START_PAGE}' is not in the allow list. Setting start page to blank."
            )

        # Define the download handler
        def _download_handler(download: Download) -> None:
            self._last_download = download

        self._download_handler = _download_handler

        # Define the Playwright controller that handles the browser interactions
        self._playwright_controller = PlaywrightController(
            animate_actions=self.animate_actions,
            downloads_folder=self.downloads_folder,
            viewport_width=self.viewport_width,
            viewport_height=self.viewport_height,
            _download_handler=self._download_handler,
            to_resize_viewport=self.to_resize_viewport,
            single_tab_mode=self.single_tab_mode,
            url_status_manager=self._url_status_manager,
            url_validation_callback=self._check_url_and_generate_msg,
        )
        self.default_tools = [
            TOOL_STOP_ACTION,
            TOOL_VISIT_URL,
            TOOL_WEB_SEARCH,
            TOOL_CLICK,
            TOOL_TYPE,
            TOOL_READ_PAGE_AND_ANSWER,
            TOOL_SLEEP,
            TOOL_HOVER,
            TOOL_HISTORY_BACK,
            TOOL_KEYPRESS,
            TOOL_REFRESH_PAGE,
            # TOOL_CLICK_FULL,
        ]
        self.did_lazy_init = False  # flag to check if we have initialized the browser
        self.is_paused = False
        self._pause_event = asyncio.Event()
        self.action_guard: BaseApprovalGuard | None = (
            ApprovalGuardContext.approval_guard() if self.use_action_guard else None
        )
        if self._model_client.model_info["vision"]:
            self.is_multimodal = True
        else:
            self.is_multimodal = False

    async def lazy_init(
        self,
    ) -> None:
        """Initialize the browser and page on first use.

        This method is called automatically on first interaction. It:
        - Starts Playwright and launches a browser
        - Creates a new browser context and page
        - Configures viewport size and downloads
        - Navigates to the start page
        - Sets up the debug directory if specified
        """
        if self.did_lazy_init:
            return

        self._last_download = None
        self._prior_metadata_hash = None

        await self._browser.__aenter__()  # 启动浏览器
        

        if isinstance(self._browser, VncDockerPlaywrightBrowser):
            self.novnc_port = self._browser.novnc_port
            self.playwright_port = self._browser.playwright_port
            
            # 启动后显示连接地址：
            logger.info(
                f"WebSurfer started with browser at `http://localhost:{self._browser.novnc_port}/vnc.html?autoconnect=true&amp;resize=scale&amp;show_dot=true&amp;scaling=local&amp;quality=7&amp;compression=0&amp;view_only=0`"
            )

        self._context = self._browser.browser_context

        # Create the page
        assert self._context is not None
        self._context.set_default_timeout(20000)  # 20 sec

        self._page = None
        self._page = await self._context.new_page()
        await self._playwright_controller.on_new_page(self._page)

        async def handle_new_page(new_pg: Page) -> None:
            # last resort on new tabs
            assert new_pg is not None
            assert self._page is not None
            await new_pg.wait_for_load_state("domcontentloaded")
            new_url = new_pg.url
            await new_pg.close()
            await self._playwright_controller.visit_page(self._page, new_url)

        if self.single_tab_mode:
            # this will make sure any new tabs will be closed and redirected to the main page
            # it is a last resort, the playwright controller handles most cases
            self._context.on("page", lambda new_pg: handle_new_page(new_pg))

        try:
            await self._playwright_controller.visit_page(self._page, self.start_page)
        except Exception:
            pass

        # Prepare the debug directory -- which stores the screenshots generated throughout the process
        await self._set_debug_dir()
        self.did_lazy_init = True

    async def pause(self) -> None:
        """Pause the WebSurfer agent."""
        self.is_paused = True
        self._pause_event.set()

    async def monitor_pause(self, cancellation_token: CancellationToken) -> None:
        """Set up background task to monitor the pause event and cancel the code execution if paused."""
        await self._pause_event.wait()
        cancellation_token.cancel()

    async def resume(self) -> None:
        """Resume the WebSurfer agent."""
        self.is_paused = False
        self._pause_event.clear()

    async def close(self) -> None:
        """
        Close the browser and the page.
        Should be called when the agent is no longer needed.
        """
        logger.info("Closing WebSurfer...")
        if self._page is not None:
            # If we're already exiting, we don't need to close the page again
            self._page = None
        await self._browser.__aexit__(None, None, None)
        if hasattr(self._model_client, "close"):
            await self._model_client.close()

    async def _set_debug_dir(self) -> None:
        assert self._page is not None
        if self.debug_dir is None:
            return
        if not os.path.isdir(self.debug_dir):
            os.mkdir(self.debug_dir)

    @property
    def produced_message_types(self) -> List[type[BaseChatMessage]]:
        """Get the types of messages produced by the agent."""
        return [MultiModalMessage]

    async def on_reset(self, cancellation_token: CancellationToken) -> None:
        """Reset the WebSurfer agent's state."""
        if not self.did_lazy_init:
            return
        assert self._page is not None

        self._chat_history.clear()
        (
            reset_prior_metadata,
            reset_last_download,
        ) = await self._playwright_controller.visit_page(self._page, self.start_page)
        if reset_last_download and self._last_download is not None:
            self._last_download = None
        if reset_prior_metadata and self._prior_metadata_hash is not None:
            self._prior_metadata_hash = None

    async def on_messages(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> Response:
        """
        Handle incoming messages and return a single response.

        Args:
            messages (Sequence[BaseChatMessage]): A sequence of incoming chat messages.
            cancellation_token (CancellationToken): Token to cancel the operation.

        Returns:
            Response: A single `Response` object generated from the incoming messages.
        """
        async for message in self.on_messages_stream(messages, cancellation_token):
            if isinstance(message, Response):
                return message
        raise AssertionError("The stream should have returned the final result.")

    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseChatMessage | Response, None]:
        """
        Handle incoming messages and yield responses as a stream.

        Args:
            messages (Sequence[BaseChatMessage]): A sequence of incoming chat messages.
            cancellation_token (CancellationToken): Token to cancel the operation.

        Yields:
            AsyncGenerator: A stream of `BaseChatMessage` or `Response` objects.
        """
        # only keep last message and any MultiModalMessages
        for i, chat_message in enumerate(messages):
            if isinstance(chat_message, MultiModalMessage):
                self._chat_history.append(
                    UserMessage(
                        content=chat_message.content, source=chat_message.source
                    )
                )
            elif isinstance(chat_message, TextMessage) and i == len(messages) - 1:
                # Only append the last TextMessage
                self._chat_history.append(
                    UserMessage(
                        content=chat_message.content, source=chat_message.source
                    )
                )
        self._last_outside_message = content_to_str(self._chat_history[-1].content)
        self.inner_messages: List[BaseChatMessage] = []
        self.model_usage: List[RequestUsage] = []
        actions_proposed: List[str] = []
        observations: List[str] = []
        action_results: List[str] = []
        emited_responses: List[str] = []
        all_screenshots: List[bytes] = []
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content="The WebSurfer is paused. Please resume to continue surfing the web.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
        non_action_tools = ["stop_action", "answer_question"]
        # first make sure the page is accessible

        assert self._page is not None

        # Set up the cancellation token for the code execution.
        llm_cancellation_token = CancellationToken()

        # Cancel the code execution if the handler's cancellation token is set.
        cancellation_token.add_callback(lambda: llm_cancellation_token.cancel())
        monitor_pause_task = asyncio.create_task(
            self.monitor_pause(llm_cancellation_token)
        )

        try:
            for _ in range(self.max_actions_per_step):
                # 1) Generate the next action to take
                if self.is_paused:
                    break
                response: str | List[FunctionCall] = ""
                need_execute_tool = False
                (
                    response,
                    rects,
                    tools,
                    element_id_mapping,
                    need_execute_tool,
                ) = await self._get_llm_response(
                    cancellation_token=llm_cancellation_token
                )
                final_usage = RequestUsage(
                    prompt_tokens=sum([u.prompt_tokens for u in self.model_usage]),
                    completion_tokens=sum(
                        [u.completion_tokens for u in self.model_usage]
                    ),
                )

                found_stop_action = False
                if not need_execute_tool:
                    summary_of_response = f"On the webpage '{str(self._page.title)}', we propose the following action: {response}"
                    self._chat_history.append(
                        AssistantMessage(
                            content=summary_of_response,
                            source=self.name,
                        )
                    )
                    assert isinstance(response, str)
                    emited_responses.append(response)
                    actions_proposed.append(summary_of_response)
                    observations.append("")
                    action_results.append("")
                    # 2 ) Emit the action proposed
                    yield Response(
                        chat_message=TextMessage(
                            content=emited_responses[-1],
                            source=self.name,
                            models_usage=final_usage,
                        ),
                        inner_messages=self.inner_messages,
                    )
                    break
                else:
                    # need to execute the tool(s)
                    assert isinstance(response, list)
                    for action in response:
                        assert isinstance(action, FunctionCall)
                        tool_call_name = action.name
                        tool_call_msg = f"{action.name}( {json.dumps(json.loads(action.arguments))} )"
                        tool_call_explanation = json.loads(action.arguments).get(
                            "explanation"
                        )
                        if tool_call_name == "answer_question":
                            tool_call_explanation = (
                                "Reading the entire page... " + tool_call_explanation
                            )
                        actions_proposed.append(tool_call_msg)
                        action_context = f'"{str(await self._page.title())}" (at {str(self._page.url)})'

                        self._chat_history.append(
                            AssistantMessage(
                                content=f"On the webpage {action_context}, we propose the following action: {tool_call_msg}",
                                source=self.name,
                            )
                        )

                        require_approval: bool = False
                        tool_args = json.loads(action.arguments)
                        assert isinstance(tool_args, dict)
                        if self.use_action_guard and self.action_guard is not None:
                            # find the tool schema given the name
                            tool_schema: ToolSchema | None = next(
                                (
                                    tool
                                    for tool in tools
                                    if tool["name"] == tool_call_name
                                ),
                                None,
                            )

                            baseline_needs_approval: MaybeRequiresApproval | None = None
                            if tool_schema is not None:
                                tool_metadata: ToolMetadata = get_tool_metadata(
                                    tool_call_name
                                )
                                baseline_needs_approval = tool_metadata.get(
                                    "requires_approval", "always"
                                )
                            else:
                                # If the tool schema is not found, default to a reasonable value
                                # In principle, this should be based on the action_guard policy
                                # e.g.: Always => Always, Never => Never, Permissive/Conservative => Maybe?
                                # Technically, Maybe always works here
                                baseline_needs_approval = "maybe"

                            assert baseline_needs_approval is not None
                            llm_guess = baseline_needs_approval  # only takes values "always" or "never"
                            # First check if require_approval is explicitly set in the tool arguments
                            # This is only for "maybe needs approval" actions
                            if "require_approval" in tool_args:
                                if tool_args["require_approval"]:
                                    llm_guess = "always"
                                else:
                                    llm_guess = "never"
                            require_approval = (
                                await self.action_guard.requires_approval(
                                    baseline_needs_approval,
                                    llm_guess,
                                    self._chat_history,
                                )
                            )

                        if tool_call_name == "stop_action":
                            tool_call_answer = json.loads(action.arguments).get(
                                "answer"
                            )
                            observations.append(tool_call_answer)
                            action_results.append(tool_call_answer)
                            emited_responses.append(tool_call_answer)
                            found_stop_action = True
                            yield Response(
                                chat_message=TextMessage(
                                    content=emited_responses[-1],
                                    source=self.name,
                                    models_usage=final_usage,
                                ),
                            )
                            break
                        if self.is_paused:
                            break
                        emited_responses.append(tool_call_explanation)
                        yield Response(
                            chat_message=TextMessage(
                                content=emited_responses[-1],
                                source=self.name,
                                models_usage=final_usage,
                            ),
                        )

                        if require_approval:
                            assert self.action_guard is not None
                            # tool_call_msg with (if exist) explanation
                            action_proposal = (
                                f"{tool_call_name}( {json.dumps(json.loads(action.arguments))} )"
                                if tool_call_explanation is None
                                else f"{tool_call_explanation}"
                            )

                            request_message = TextMessage(
                                source=self.name,
                                content=f"On the webpage {action_context}, we propose the following action: {action_proposal}",
                            )
                            # Preview action for tools that have target_id
                            if tool_call_name in [
                                "click",
                                "hover",
                                "select_option",
                                "upload_file",
                            ]:
                                if "target_id" in tool_args:
                                    target_id = str(tool_args["target_id"])  # type: ignore
                                    assert isinstance(target_id, str)
                                    if target_id in element_id_mapping:
                                        await (
                                            self._playwright_controller.preview_action(
                                                self._page,
                                                element_id_mapping[target_id],
                                            )
                                        )
                            elif (
                                tool_call_name == "input_text"
                                and "input_field_id" in tool_args
                            ):
                                input_field_id = str(tool_args["input_field_id"])  # type: ignore
                                assert isinstance(input_field_id, str)
                                if input_field_id in element_id_mapping:
                                    await self._playwright_controller.preview_action(
                                        self._page, element_id_mapping[input_field_id]
                                    )

                            approval = await self.action_guard.get_approval(
                                request_message,
                            )

                            if not approval:
                                # cleanup animations
                                await self._playwright_controller.cleanup_animations(
                                    self._page
                                )
                                # User did not approve the action
                                self._chat_history.append(
                                    AssistantMessage(
                                        content="The action was not approved.",
                                        source=self.name,
                                    )
                                )
                                yield Response(
                                    chat_message=TextMessage(
                                        content="The action was not approved.",
                                        source=self.name,
                                        models_usage=final_usage,
                                    ),
                                )
                                break

                        # execute the tool call
                        action_result = ""
                        try:
                            action_result = await self._execute_tool(
                                message=[action],
                                rects=rects,
                                tools=tools,
                                element_id_mapping=element_id_mapping,
                                cancellation_token=llm_cancellation_token,
                            )
                        except RuntimeError as e:
                            if "WebSurfer was paused" in str(e):
                                self.logger.info(f"Tool execution paused: {e}")
                                action_result = "Action was interrupted because WebSurfer was paused."
                            else:
                                self.logger.error(f"Runtime error executing tool: {e}")
                                action_result = f"Error occurred while executing action {tool_call_msg}: {e}"
                        except Exception as e:
                            self.logger.error(f"Error executing tool: {e}")
                            action_result = f"Error occurred while executing action {tool_call_msg}: {e}"
                        new_screenshot = (
                            await self._playwright_controller.get_screenshot(self._page)
                        )
                        if self.to_save_screenshots and self.debug_dir is not None:
                            current_timestamp = "_" + int(time.time()).__str__()
                            screenshot_png_name = (
                                "screenshot_raw" + current_timestamp + ".png"
                            )
                            PIL.Image.open(io.BytesIO(new_screenshot)).save(
                                os.path.join(self.debug_dir, screenshot_png_name)
                            )
                        all_screenshots.append(new_screenshot)
                        content: list[str | AGImage] = [
                            action_result,
                            AGImage.from_pil(
                                PIL.Image.open(io.BytesIO(new_screenshot))
                            ),
                        ]
                        emited_responses.append(action_result)
                        # 4) Emit the observation
                        yield Response(
                            chat_message=MultiModalMessage(
                                content=content,
                                source=self.name,
                                models_usage=final_usage,
                                metadata={
                                    "internal": "no",
                                    "type": "browser_screenshot",
                                },
                            ),
                            inner_messages=self.inner_messages,
                        )
                        (
                            message_content,
                            _,
                            metadata_hash,
                        ) = await self._playwright_controller.describe_page(
                            self._page,
                            get_screenshot=False,
                        )
                        observations.append(f"{action_result}\n\n{message_content}")
                        action_results.append(action_result)

                        self._chat_history.append(
                            UserMessage(
                                content=[
                                    f"Observation: {action_result}\n\n{message_content}",
                                    AGImage.from_pil(
                                        PIL.Image.open(io.BytesIO(new_screenshot))
                                    ),
                                ],
                                source=self.name,
                            )
                        )
                        if tool_call_name in non_action_tools:
                            found_stop_action = True
                            break
                    if found_stop_action:
                        break
        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=self.inner_messages,
            )
        except Exception as e:
            self.logger.error(f"Error in on_messages: {e}")
            pass
        finally:
            # Cancel the monitor task.
            monitor_pause_task.cancel()
            try:
                await monitor_pause_task
            except asyncio.CancelledError:
                pass

        # Prepare Final Response to other agents

        all_responses = (
            "The actions the websurfer performed are the following.\n"
            + "\n".join(
                [
                    f"\n Action: {action}\nObservation: {action_result}\n\n"
                    for action, action_result in zip(actions_proposed, action_results)
                ]
            )
        )
        try:
            (
                message_content,
                maybe_new_screenshot,
                metadata_hash,
            ) = await self._playwright_controller.describe_page(self._page)
            self._prior_metadata_hash = metadata_hash

            message_content = f"\n\n{all_responses}\n\n" + message_content
            assert maybe_new_screenshot is not None
            new_screenshot = maybe_new_screenshot

            content = [
                message_content,
                AGImage.from_pil(PIL.Image.open(io.BytesIO(new_screenshot))),
            ]

            final_usage = RequestUsage(
                prompt_tokens=sum([u.prompt_tokens for u in self.model_usage]),
                completion_tokens=sum([u.completion_tokens for u in self.model_usage]),
            )
            # Send the final response to other agents
            yield Response(
                chat_message=MultiModalMessage(
                    content=content,
                    source=self.name,
                    models_usage=final_usage,
                    metadata={"internal": "yes"},
                ),
                inner_messages=self.inner_messages,
            )
        except Exception as e:
            self.logger.error(f"Error in on_messages: {e}")
            final_message = f"{all_responses}\n\n Encountered an error while getting screenshot: {e}"
            yield Response(
                chat_message=TextMessage(
                    content=final_message,
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )

    @staticmethod
    def _tools_to_names(tools: List[ToolSchema]) -> str:
        """Convert the list of tools to a string of names.

        Returns:
            str: A string of tool names separated by commas.
        """
        return "\n".join([t["name"] for t in tools])

    async def _get_llm_response(
        self, cancellation_token: Optional[CancellationToken] = None
    ) -> tuple[
        Union[str, List[FunctionCall]],
        Dict[str, InteractiveRegion],
        List[ToolSchema],
        Dict[str, str],
        bool,
    ]:
        """Generate the next action to take based on the current page state.

        Args:
            cancellation_token (CancellationToken, optional): Token to cancel the operation. Default: None

        Returns:
            Tuple containing:
                - str | List[FunctionCall]: The model's response (text or function calls)
                - Dict[str, InteractiveRegion]: Dictionary of interactive page elements
                - List[ToolSchema]: String of available tool names
                - Dict[str, str]: Mapping of element IDs
                - bool: Boolean indicating if tool execution is needed
        """

        # Lazy init, initialize the browser and the page on the first generate reply only
        if not self.did_lazy_init:
            await self.lazy_init()

        try:
            assert self._page is not None
            assert (
                await self._playwright_controller.get_interactive_rects(self._page)
                is not None
            )
        except Exception as e:
            # open a new tab and point it to about:blank
            self.logger.error(f"Page is not accessible, creating a new one: {e}")
            assert self._context is not None
            self._page = await self._playwright_controller.create_new_tab(
                self._context, "about:blank"
            )

        # Clone the messages to give context, removing old screenshots
        history: List[LLMMessage] = []
        date_today = datetime.now().strftime("%Y-%m-%d")
        history.append(
            SystemMessage(
                content=WEB_SURFER_SYSTEM_MESSAGE.format(date_today=date_today)
            )
        )
        # Keep images only for user messages, remove from others
        filtered_history: List[LLMMessage] = []
        for msg in self._chat_history:
            if isinstance(msg, UserMessage) and msg.source in ["user", "user_proxy"]:
                filtered_history.append(msg)
            else:
                filtered_history.extend(remove_images([msg]))
        history.extend(filtered_history)

        # Ask the page for interactive elements, then prepare the state-of-mark screenshot
        rects = await self._playwright_controller.get_interactive_rects(self._page)
        viewport = await self._playwright_controller.get_visual_viewport(self._page)
        screenshot = await self._playwright_controller.get_screenshot(self._page)
        som_screenshot, visible_rects, rects_above, rects_below, element_id_mapping = (
            add_set_of_mark(screenshot, rects, use_sequential_ids=True)
        )
        # element_id_mapping is a mapping of new ids to original ids in the page
        # we need to reverse it to get the original ids from the new ids
        # for each element we click, we need to use the original id
        reverse_element_id_mapping = {v: k for k, v in element_id_mapping.items()}
        rects = {reverse_element_id_mapping.get(k, k): v for k, v in rects.items()}

        if self.to_save_screenshots and self.debug_dir is not None:
            current_timestamp = "_" + int(time.time()).__str__()
            screenshot_png_name = "screenshot_som" + current_timestamp + ".png"
            som_screenshot.save(os.path.join(self.debug_dir, screenshot_png_name))
            self.logger.debug(
                WebSurferEvent(
                    source=self.name,
                    url=self._page.url,
                    message="Screenshot: " + screenshot_png_name,
                )
            )

        # Get the tabs information
        tabs_information_str = ""
        num_tabs = 1
        if not self.single_tab_mode and self._context is not None:
            num_tabs, tabs_information_str = await self.get_tabs_info()
            tabs_information_str = f"There are {num_tabs} tabs open. The tabs are as follows:\n{tabs_information_str}"

        # What tools are available?
        tools = self.default_tools.copy()

        # If not in single tab mode, always allow creating new tabs
        if not self.single_tab_mode:
            if TOOL_CREATE_TAB not in tools:
                tools.append(TOOL_CREATE_TAB)

        # If there are multiple tabs, we can switch between them and close them
        if not self.single_tab_mode and num_tabs > 1:
            tools.append(TOOL_SWITCH_TAB)
            tools.append(TOOL_CLOSE_TAB)

        # We can scroll up
        if viewport["pageTop"] > 5:
            tools.append(TOOL_PAGE_UP)

        # Can scroll down
        if (viewport["pageTop"] + viewport["height"] + 5) < viewport["scrollHeight"]:
            tools.append(TOOL_PAGE_DOWN)

        # Add select_option tool only if there are option elements
        if any(rect.get("role") == "option" for rect in rects.values()):
            tools.append(TOOL_SELECT_OPTION)

        # Add upload_file tool only if there are file input elements
        # if any(rect.get("tag_name") == "input, type=file" for rect in rects.values()):
        #    tools.append(TOOL_UPLOAD_FILE)

        # Focus hint
        focused = await self._playwright_controller.get_focused_rect_id(self._page)
        focused = reverse_element_id_mapping.get(focused, focused)

        focused_hint = ""
        if focused:
            name = self._target_name(focused, rects)
            if name:
                name = f"(and name '{name}') "

            role = "control"
            try:
                role = rects[focused]["role"]
            except KeyError:
                pass

            focused_hint = f"\nThe {role} with ID {focused} {name}currently has the input focus.\n\n"

        # Everything visible
        visible_targets = (
            "\n".join(self._format_target_list(visible_rects, rects)) + "\n\n"
        )

        # Everything else
        other_targets: List[str] = []
        other_targets.extend(self._format_target_list(rects_above, rects))
        other_targets.extend(self._format_target_list(rects_below, rects))

        if len(other_targets) > 0:
            # Extract just the names from the JSON strings
            other_target_names: List[str] = []
            for target in other_targets:
                try:
                    target_dict = json.loads(target)
                    name = target_dict.get("name", "")
                    role = target_dict.get("role", "")
                    other_target_names.append(name if name else f"{role} control")
                except json.JSONDecodeError:
                    continue

            other_targets_str = (
                "Some additional valid interaction targets (not shown, you need to scroll to interact with them) include:\n"
                + ", ".join(other_target_names[:30])
                + "\n\n"
            )
        else:
            other_targets_str = ""

        tool_names = WebSurfer._tools_to_names(tools)

        webpage_text = await self._playwright_controller.get_visible_text(self._page)

        if not self.json_model_output:
            text_prompt = WEB_SURFER_TOOL_PROMPT.format(
                tabs_information=tabs_information_str,
                last_outside_message=self._last_outside_message,
                webpage_text=webpage_text,
                url=self._page.url,
                visible_targets=visible_targets,
                consider_screenshot="Consider the following screenshot of a web browser,"
                if self.is_multimodal
                else "Consider the following webpage",
                other_targets_str=other_targets_str,
                focused_hint=focused_hint,
                tool_names=tool_names,
            ).strip()
        else:
            text_prompt = WEB_SURFER_NO_TOOLS_PROMPT.format(
                tabs_information=tabs_information_str,
                last_outside_message=self._last_outside_message,
                webpage_text=webpage_text,
                url=self._page.url,
                visible_targets=visible_targets,
                consider_screenshot="Consider the following screenshot of a web browser,"
                if self.is_multimodal
                else "Consider the following webpage",
                other_targets_str=other_targets_str,
                focused_hint=focused_hint,
            ).strip()

        if self.is_multimodal:
            # Scale the screenshot for the MLM, and close the original
            scaled_som_screenshot = som_screenshot.resize(
                (self.MLM_WIDTH, self.MLM_HEIGHT)
            )
            screenshot_file = PIL.Image.open(io.BytesIO(screenshot))
            scaled_screenshot = screenshot_file.resize(
                (self.MLM_WIDTH, self.MLM_HEIGHT)
            )
            som_screenshot.close()
            screenshot_file.close()

            # Add the multimodal message and make the request
            history.append(
                UserMessage(
                    content=[
                        text_prompt,
                        AGImage.from_pil(scaled_som_screenshot),
                        AGImage.from_pil(scaled_screenshot),
                    ],
                    source=self.name,
                )
            )
        else:
            history.append(
                UserMessage(
                    content=text_prompt,
                    source=self.name,
                )
            )

        # Re-initialize model context to meet token limit quota
        try:
            await self._model_context.clear()
            for msg in history:
                await self._model_context.add_message(msg)
            token_limited_history = await self._model_context.get_messages()
        except Exception:
            token_limited_history = history

        if not self.json_model_output:
            create_args: Dict[str, Any] | None = None
            if self._model_client.model_info["family"] in [
                "gpt-4o",
                "gpt-41",
                "gpt-45",
                "o3",
                "o4",
            ]:
                create_args = {
                    "tool_choice": "required",
                }
                if self.multiple_tools_per_call:
                    create_args["parallel_tool_calls"] = True
            if create_args is not None:
                response = await self._model_client.create(
                    token_limited_history,
                    tools=tools,
                    cancellation_token=cancellation_token,
                    extra_create_args=create_args,
                )
            else:
                response = await self._model_client.create(
                    token_limited_history,
                    tools=tools,
                    cancellation_token=cancellation_token,
                )
        else:
            response = await self._model_client.create(
                token_limited_history,
                cancellation_token=cancellation_token,
            )
        self.model_usage.append(response.usage)
        self._last_download = None
        if not self.json_model_output:
            to_execute_tool = isinstance(response.content, list)
            return (
                response.content,
                rects,
                tools,
                element_id_mapping,
                to_execute_tool,
            )
        else:
            try:
                # check if first line is `json
                response_content = response.content
                assert isinstance(response_content, str)
                if response_content.startswith("```json"):
                    # remove first and last line
                    response_lines = response_content.split("\n")
                    response_lines = response_lines[1:-1]
                    response_content = "\n".join(response_lines)

                json_response = json.loads(response_content)
                tool_name = json_response["tool_name"]
                tool_args = json_response["tool_args"]
                tool_args["explanation"] = json_response["explanation"]
                function_call = FunctionCall(
                    id="json_response", name=tool_name, arguments=json.dumps(tool_args)
                )
                return [function_call], rects, tools, element_id_mapping, True
            except Exception as e:
                error_msg = f"Failed to parse JSON response: {str(e)}. Response was: {response.content}"
                return error_msg, rects, tools, element_id_mapping, False

    async def _check_url_and_generate_msg(self, url: str) -> Tuple[str, bool]:
        """Returns a message to caller if the URL is not allowed and a boolean indicating if the user has approved the URL."""
        # TODO: Hacky check to see if the URL was aborted. Find a better way to do this
        if url == "chrome-error://chromewebdata/":
            if self._last_rejected_url is not None:
                last_rejected_url = self._last_rejected_url
                self._last_rejected_url = None
                return (
                    f"I am not allowed to access the website {last_rejected_url} because it is not in the list of websites I can access and the user has declined to approve it.",
                    False,
                )

        if self._url_status_manager.is_url_blocked(url):
            return (
                f"I am not allowed to access the website {url} because has been blocked.",
                False,
            )

        if not self._url_status_manager.is_url_allowed(url):
            if not self._url_status_manager.is_url_rejected(url):
                # tldextract will only recombine entries with valid, registered hostnames. We'll just use the straight url for anything else.
                domain = tldextract.extract(url).fqdn
                if not domain:
                    domain = url
                response = False
                if self.action_guard is not None:
                    # The content string here is important because the UI is checking for specific wording to detect the URL approval request
                    request_message = TextMessage(
                        source=self.name,
                        content=f"The website {url} is not allowed. Would you like to allow the domain {domain} for this session?",
                    )
                    response = await self.action_guard.get_approval(request_message)
                if response:
                    self._url_status_manager.set_url_status(domain, URL_ALLOWED)
                    return (
                        "",
                        True,
                    )
                else:
                    self._url_status_manager.set_url_status(domain, URL_REJECTED)

            self._last_rejected_url = url
            return (
                f"I am not allowed to access the website {url} because it is not in the list of websites I can access and the user has declined to allow it.",
                False,
            )
        return "", True

    async def _execute_tool_stop_action(self, args: Dict[str, Any]) -> str:
        answer = cast(str, args.get("answer"))
        return answer

    async def _execute_tool_visit_url(self, args: Dict[str, Any]) -> str:
        url = cast(str, args.get("url"))
        ret, approved = await self._check_url_and_generate_msg(url)
        if not approved:
            return ret

        action_description = f"I typed '{url}' into the browser address bar."
        assert self._page is not None
        if url.startswith(("https://", "http://", "file://", "about:")):
            (
                reset_prior_metadata,
                reset_last_download,
            ) = await self._playwright_controller.visit_page(self._page, url)
        # If the argument contains a space, treat it as a search query
        elif " " in url:
            ret, approved = await self._check_url_and_generate_msg("bing.com")
            if not approved:
                return ret
            (
                reset_prior_metadata,
                reset_last_download,
            ) = await self._playwright_controller.visit_page(
                self._page,
                f"https://www.bing.com/search?q={quote_plus(url)}&FORM=QBLH",
            )
        # Otherwise, prefix with https://
        else:
            (
                reset_prior_metadata,
                reset_last_download,
            ) = await self._playwright_controller.visit_page(
                self._page, "https://" + url
            )

        if reset_last_download:
            self._last_download = None
        if reset_prior_metadata:
            self._prior_metadata_hash = None

        return action_description

    async def _execute_tool_history_back(self, args: Dict[str, Any]) -> str:
        assert self._page is not None
        response = await self._playwright_controller.go_back(self._page)
        if response:
            return "I clicked the browser back button."
        return "No previous page in the browser history or couldn't navigate back."

    async def _execute_tool_refresh_page(self, args: Dict[str, Any]) -> str:
        assert self._page is not None
        await self._playwright_controller.refresh_page(self._page)
        return "I refreshed the current page."

    async def _execute_tool_web_search(self, args: Dict[str, Any]) -> str:
        assert self._page is not None
        ret, approved = await self._check_url_and_generate_msg("bing.com")
        if not approved:
            return ret
        query = cast(str, args.get("query"))
        action_description = f"I typed '{query}' into the browser search bar."
        (
            reset_prior_metadata,
            reset_last_download,
        ) = await self._playwright_controller.visit_page(
            self._page,
            f"https://www.bing.com/search?q={quote_plus(query)}&FORM=QBLH",
        )
        if reset_last_download:
            self._last_download = None
        if reset_prior_metadata:
            self._prior_metadata_hash = None
        return action_description

    async def _execute_tool_page_up(self, args: Dict[str, Any]) -> str:
        assert self._page is not None
        await self._playwright_controller.page_up(self._page)
        return "I scrolled up one page in the browser."

    async def _execute_tool_page_down(self, args: Dict[str, Any]) -> str:
        assert self._page is not None
        await self._playwright_controller.page_down(self._page)
        return "I scrolled down one page in the browser."

    async def _execute_tool_click(
        self,
        args: Dict[str, Any],
        rects: Dict[str, InteractiveRegion],
        element_id_mapping: Dict[str, str],
    ) -> str:
        assert "target_id" in args
        target_id: str = str(args["target_id"])
        target_name = self._target_name(target_id, rects)
        target_id = element_id_mapping[target_id]

        action_description = (
            f"I clicked '{target_name}'." if target_name else "I clicked the control."
        )
        assert self._context is not None
        assert self._page is not None
        new_page_tentative = await self._playwright_controller.click_id(
            self._context, self._page, target_id
        )
        if new_page_tentative is not None:
            new_url = new_page_tentative.url
            self._page = new_page_tentative
            self._prior_metadata_hash = None
            ret, approved = await self._check_url_and_generate_msg(new_url)
            if not approved:
                return ret

        return action_description

    async def _execute_tool_click_full(
        self,
        args: Dict[str, Any],
        rects: Dict[str, InteractiveRegion],
        element_id_mapping: Dict[str, str],
    ) -> str:
        assert "target_id" in args
        target_id: str = str(args["target_id"])
        target_name = self._target_name(target_id, rects)
        target_id = element_id_mapping[target_id]

        hold = float(args.get("hold", 0.0))
        button = str(args.get("button", "left"))
        button = cast(Literal["left", "right"], button)
        action_description = (
            f"I clicked '{target_name}' with button '{button}' and hold {hold} seconds."
            if target_name
            else f"I clicked the control with button '{button}' and hold {hold} seconds."
        )
        assert self._context is not None
        assert self._page is not None
        new_page_tentative = await self._playwright_controller.click_id(
            self._context, self._page, target_id, hold=hold, button=button
        )
        if new_page_tentative is not None:
            new_url = new_page_tentative.url
            self._page = new_page_tentative
            self._prior_metadata_hash = None
            ret, approved = await self._check_url_and_generate_msg(new_url)
            if not approved:
                return ret

        return action_description

    async def _execute_tool_input_text(
        self,
        args: Dict[str, Any],
        rects: Dict[str, InteractiveRegion],
        element_id_mapping: Dict[str, str],
    ) -> str:
        assert "input_field_id" in args
        assert "text_value" in args
        assert "press_enter" in args
        assert "delete_existing_text" in args
        input_field_id: str = str(args["input_field_id"])
        input_field_name = self._target_name(input_field_id, rects)
        input_field_id = element_id_mapping[input_field_id]

        text_value = str(args.get("text_value"))
        press_enter = bool(args.get("press_enter"))
        delete_existing_text = bool(args.get("delete_existing_text"))

        action_description = (
            f"I typed '{text_value}' into '{input_field_name}'."
            if input_field_name
            else f"I typed '{text_value}'."
        )
        assert self._page is not None
        await self._playwright_controller.fill_id(
            self._page,
            input_field_id,
            text_value,
            press_enter,
            delete_existing_text,
        )
        return action_description

    async def _execute_tool_answer_question(
        self,
        args: Dict[str, Any],
        cancellation_token: Optional[CancellationToken] = None,
    ) -> str:
        assert "question" in args
        question: str = args["question"]
        return await self._summarize_page(
            question=question, cancellation_token=cancellation_token
        )

    async def _execute_tool_summarize_page(
        self,
        args: Dict[str, Any],
        cancellation_token: Optional[CancellationToken] = None,
    ) -> str:
        return await self._summarize_page(cancellation_token=cancellation_token)

    async def _execute_tool_hover(
        self,
        args: Dict[str, Any],
        rects: Dict[str, InteractiveRegion],
        element_id_mapping: Dict[str, str],
    ) -> str:
        target_id = str(args.get("target_id"))
        target_name = self._target_name(target_id, rects)
        target_id = element_id_mapping.get(target_id, target_id)

        action_description = (
            f"I hovered over '{target_name}'."
            if target_name
            else "I hovered over the control."
        )
        assert self._page is not None
        await self._playwright_controller.hover_id(self._page, target_id)
        return action_description

    async def _execute_tool_sleep(self, args: Dict[str, Any]) -> str:
        assert self._page is not None
        duration = cast(int, args.get("duration", 3))
        await self._playwright_controller.sleep(self._page, duration)
        return f"I waited {duration} seconds."

    async def _execute_tool_select_option(
        self,
        args: Dict[str, Any],
        rects: Dict[str, InteractiveRegion],
        element_id_mapping: Dict[str, str],
    ) -> str:
        target_id = str(args.get("target_id"))
        target_name = self._target_name(target_id, rects)
        target_id = element_id_mapping.get(target_id, target_id)

        action_description = (
            f"I selected the option '{target_name}'."
            if target_name
            else "I selected the option."
        )
        assert self._context is not None
        assert self._page is not None
        new_page_tentative = await self._playwright_controller.select_option(
            self._context, self._page, target_id
        )
        if new_page_tentative is not None:
            new_url = new_page_tentative.url
            ret, approved = await self._check_url_and_generate_msg(new_url)
            if not approved:
                return ret
            self._page = new_page_tentative
            self._prior_metadata_hash = None

        return action_description

    async def _execute_tool_create_tab(self, args: Dict[str, Any]) -> str:
        url: str = cast(str, args.get("url"))
        ret, approved = await self._check_url_and_generate_msg(url)
        if not approved:
            return ret
        assert self._context is not None
        action_description = f"I created a new tab and navigated to '{url}'."
        new_page = await self._playwright_controller.create_new_tab(self._context, url)
        self._page = new_page
        self._prior_metadata_hash = None
        return action_description

    async def _execute_tool_switch_tab(self, args: Dict[str, Any]) -> str:
        assert self._context is not None
        try:
            assert "tab_index" in args
            tab_index = int(cast(str, args["tab_index"]))
            new_page = await self._playwright_controller.switch_tab(
                self._context, tab_index
            )
            self._page = new_page
            self._prior_metadata_hash = None
            return f"I switched to tab {tab_index}."
        except (ValueError, TypeError) as e:
            return f"Invalid tab index: {e}"

    async def _execute_tool_close_tab(self, args: Dict[str, Any]) -> str:
        try:
            assert self._context is not None
            assert "tab_index" in args
            tab_index = int(cast(str, args["tab_index"]))
            new_page = await self._playwright_controller.close_tab(
                self._context, tab_index
            )
            self._page = new_page
            self._prior_metadata_hash = None
            return f"I closed tab {tab_index} and switched to an adjacent tab."
        except (ValueError, TypeError) as e:
            return f"Invalid tab index: {e}"

    async def _execute_tool_upload_file(
        self,
        args: Dict[str, Any],
        rects: Dict[str, InteractiveRegion],
        element_id_mapping: Dict[str, str],
    ) -> str:
        target_id = str(args.get("target_id"))
        assert self.downloads_folder is not None
        file_path = self.downloads_folder + "/" + str(args.get("file_path"))
        target_name = self._target_name(target_id, rects)

        target_id = element_id_mapping.get(target_id, target_id)
        if target_name:
            action_description = (
                f"I uploaded the file '{file_path}' to '{target_name}'."
            )
        else:
            action_description = f"I uploaded the file '{file_path}'."
        assert self._page is not None
        await self._playwright_controller.upload_file(self._page, target_id, file_path)

        return action_description

    async def _execute_tool_keypress(self, args: Dict[str, Any]) -> str:
        """Execute the keypress tool to press one or more keyboard keys in sequence.

        Args:
            args (Dict[str, Any]): The arguments for the keypress tool.
                keys (List[str]): List of keys to press in sequence.

        Returns:
            str: A message describing what was done.
        """
        assert self._page is not None, "Browser page is not initialized"
        keys = args["keys"]
        await self._playwright_controller.keypress(self._page, keys)
        keys_str = ", ".join(f"'{k}'" for k in keys)
        return f"Pressed the following keys in sequence: {keys_str}"

    async def _execute_tool(
        self,
        message: List[FunctionCall],
        rects: Dict[str, InteractiveRegion],
        tools: List[ToolSchema],
        element_id_mapping: Dict[str, str],
        cancellation_token: Optional[CancellationToken] = None,
    ) -> str:
        """Execute a tool action in the browser.

        Args:
            message (List[FunctionCall]): List of function calls from the model, should be a single function call
            rects (Dict[str, InteractiveRegion]): Dictionary of interactive page elements
            tools (List[ToolSchema]): List of available tool schemas
            element_id_mapping (Dict[str, str]): Mapping of element IDs
            cancellation_token (CancellationToken, optional): Token to cancel the operation. Default: None

        Returns:
            str: Description of the action taken

        Raises:
            ValueError: If an unknown tool is specified
            RuntimeError: If the WebSurfer was paused during tool execution
        """
        assert self._context is not None, "Browser context is not initialized"
        assert len(message) == 1, "Expected exactly one function call"
        assert self._page is not None

        name = message[0].name
        args = json.loads(message[0].arguments)

        self.logger.debug(
            WebSurferEvent(
                source=self.name,
                url=self._page.url,
                action=name,
                arguments=args,
                message=f"{name}( {json.dumps(args)} )",
            )
        )
        self.inner_messages.append(
            TextMessage(content=f"{name}( {json.dumps(args)} )", source=self.name)
        )

        # Convert tool name to function name (e.g. "visit_url" -> "_execute_tool_visit_url")
        tool_func_name = f"_execute_tool_{name}"
        tool_func = getattr(self, tool_func_name, None)

        if tool_func is None:
            tool_names = WebSurfer._tools_to_names(tools)

            raise ValueError(
                f"Unknown tool '{name}'. Please choose from:\n\n{tool_names}"
            )

        # Create the appropriate arguments based on the tool's requirements
        tool_kwargs = {"args": args}
        if name in [
            "click",
            "input_text",
            "hover",
            "select_option",
            "upload_file",
            "click_full",
        ]:
            tool_kwargs.update(
                {"rects": rects, "element_id_mapping": element_id_mapping}
            )
        if name in ["answer_question", "summarize_page"]:
            tool_kwargs["cancellation_token"] = cancellation_token

        # Start a new task to execute the tool.
        execute_tool_task = asyncio.create_task(tool_func(**tool_kwargs))

        # Start a new task to wait for the pause event and cancel the tool execution if paused.
        async def wait_for_pause_event() -> None:
            await self._pause_event.wait()
            if self._pause_event.is_set():
                execute_tool_task.cancel()

        wait_for_pause_task = asyncio.create_task(wait_for_pause_event())

        # Wait for the tool execution to complete or be cancelled.
        try:
            action_description = await execute_tool_task
            assert isinstance(action_description, str)
            # Handle downloads
            if self._last_download is not None and self.downloads_folder is not None:
                action_description += f"\n\nSuccessfully downloaded '{self._last_download.suggested_filename}' to local path: {self.downloads_folder}"
        except asyncio.CancelledError:
            # If the task was cancelled, we can handle it here if needed
            # Clean up any cursor animations or highlights that were added by animate_actions.
            assert self._page is not None
            await self._playwright_controller.cleanup_animations(self._page)
            return f"WebSurfer was paused, action '{name}' was cancelled."
        finally:
            # Cancel the wait_for_pause_task if it is still running
            if not wait_for_pause_task.done():
                wait_for_pause_task.cancel()
        # cleanup animations
        await self._playwright_controller.cleanup_animations(self._page)
        return action_description

    def _target_name(
        self, target: str, rects: Dict[str, InteractiveRegion]
    ) -> str | None:
        """Get the accessible name of a target element.

        Args:
            target (str): ID of the target element
            rects (Dict[str, InteractiveRegion]): Dictionary of interactive page elements

        Returns:
            str | None: The aria name of the element if available, None otherwise
        """
        try:
            return rects[target]["aria_name"].strip()
        except KeyError:
            return None

    def _format_target_list(
        self, ids: List[str], rects: Dict[str, InteractiveRegion]
    ) -> List[str]:
        """
        Format the list of targets in the webpage as a string to be used in the agent's prompt.
        """
        targets: List[str] = []
        for r in list(set(ids)):
            if r in rects:
                # Get the role
                aria_role = rects[r].get("role", "").strip()
                if len(aria_role) == 0:
                    aria_role = rects[r].get("tag_name", "").strip()

                # Get the name
                aria_name = re.sub(
                    r"[\n\r]+", " ", rects[r].get("aria_name", "")
                ).strip()

                # What are the actions?
                actions = ['"click", "hover"']
                if (
                    rects[r]["role"] in ["textbox", "searchbox", "combobox"]
                    or rects[r].get("tag_name") in ["input", "textarea", "search"]
                    or rects[r].get("contenteditable") == "true"
                ):
                    actions.append('"input_text"')
                # if the role is option, add "select" to the actions
                if rects[r]["role"] == "option":
                    actions = ['"select_option"']
                # check if the role is file input
                if aria_role == "input, type=file":
                    actions = ['"upload_file"']
                actions_str = "[" + ",".join(actions) + "]"
                # limit  name to maximum 100 characters
                aria_name = aria_name[:100]
                targets.append(
                    f'{{"id": {r}, "name": "{aria_name}", "role": "{aria_role}", "tools": {actions_str} }}'
                )
        sorted_targets = sorted(
            targets, key=lambda x: int(x.split(",")[0].split(":")[1])
        )

        return sorted_targets

    async def _get_ocr_text(
        self,
        image: Union[bytes, io.BufferedIOBase, PIL.Image.Image],
        cancellation_token: Optional[CancellationToken] = None,
    ) -> str:
        """Extract text from an image using OCR via the LLM.

        Args:
            image (bytes | io.BufferedIOBase | PIL.Image.Image): Image data as bytes, file or PIL Image
            cancellation_token (CancellationToken, optional): Token to cancel the operation. Default: None

        Returns:
            str: Extracted text from the image
        """
        scaled_screenshot = None
        if isinstance(image, PIL.Image.Image):
            scaled_screenshot = image.resize((self.MLM_WIDTH, self.MLM_HEIGHT))
        else:
            pil_image = None
            if not isinstance(image, io.BufferedIOBase):
                pil_image = PIL.Image.open(io.BytesIO(image))
            else:
                pil_image = PIL.Image.open(cast(BinaryIO, image))
            scaled_screenshot = pil_image.resize((self.MLM_WIDTH, self.MLM_HEIGHT))
            pil_image.close()

        # Add the multimodal message and make the request
        messages: List[LLMMessage] = []
        messages.append(
            UserMessage(
                content=[
                    WEB_SURFER_OCR_PROMPT,
                    AGImage.from_pil(scaled_screenshot),
                ],
                source=self.name,
            )
        )
        response = await self._model_client.create(
            messages, cancellation_token=cancellation_token
        )
        self.model_usage.append(response.usage)
        scaled_screenshot.close()
        assert isinstance(response.content, str)
        return response.content

    async def _summarize_page(
        self,
        question: Optional[str] = None,
        cancellation_token: Optional[CancellationToken] = None,
    ) -> str:
        """Generate a summary of the current page content.

        Args:
            question (str, optional): Optional specific question to answer about the page. Default: None
            cancellation_token (CancellationToken, optional): Token to cancel the operation. Default: None

        Returns:
            str: Summary text or answer to the question
        """
        assert self._page is not None

        # Get page content and title
        page_markdown: str = await self._playwright_controller.get_page_markdown(
            self._page
        )
        title: str = await self._page.title() or self._page.url

        # Take a screenshot and scale it
        screenshot = await self._playwright_controller.get_screenshot(self._page)
        pil_image = PIL.Image.open(io.BytesIO(screenshot))
        scaled_screenshot = pil_image.resize((self.MLM_WIDTH, self.MLM_HEIGHT))
        pil_image.close()
        ag_image = AGImage.from_pil(scaled_screenshot)

        # Prepare the system prompt and user prompt
        messages: List[LLMMessage] = []
        messages.append(SystemMessage(content=WEB_SURFER_QA_SYSTEM_MESSAGE))

        # Create the prompt with the question
        prompt = WEB_SURFER_QA_PROMPT(title, question)

        # Truncate the page content if needed to fit within token limits
        tokenizer = tiktoken.encoding_for_model("gpt-4o")
        prompt_tokens = len(tokenizer.encode(prompt))
        # Reserve tokens for the image (SCREENSHOT_TOKENS) and some buffer for the response
        max_content_tokens = 128000 - self.SCREENSHOT_TOKENS - prompt_tokens - 1000

        if max_content_tokens <= 0:
            # If we don't have enough tokens, just use a minimal prompt
            content = prompt
        else:
            # Truncate the page content to fit within the token limit
            content_tokens = tokenizer.encode(page_markdown)
            if len(content_tokens) > max_content_tokens:
                truncated_content = tokenizer.decode(
                    content_tokens[:max_content_tokens]
                )
                content = f"Page content (truncated):\n{truncated_content}\n\n{prompt}"
            else:
                content = f"Page content:\n{page_markdown}\n\n{prompt}"

        # Create the message with the content and image
        messages.append(
            UserMessage(
                content=[content, ag_image],
                source=self.name,
            )
        )

        # Generate the response
        response = await self._model_client.create(
            messages, cancellation_token=cancellation_token
        )
        self.model_usage.append(response.usage)
        scaled_screenshot.close()

        assert isinstance(response.content, str)
        return response.content

    async def describe_current_page(self) -> tuple[str, Union[bytes, None], str]:
        """Get a description of the current page including content, screenshot and metadata hash.

        Returns:
            Tuple containing:
            - str: String description of the page content
            - bytes | None: Screenshot of the page as bytes
            - str: Hash of the page metadata
        """
        assert self._page is not None
        return await self._playwright_controller.describe_page(self._page)

    async def get_page_title_url(self) -> tuple[str, str]:
        """Get the title and URL of the current page.

        Returns:
            Tuple containing:
            - str: String title of the page
            - str: String URL of the page
        """
        assert self._page is not None
        return await self._page.title(), self._page.url

    async def get_tabs_info(self) -> Tuple[int, str]:
        """Returns the number of tabs and a newline delineated string describing each of them. An example of the string is:

        Tab 0: <Tab_Title> (<URL>) [CURRENTLY SHOWN] [CONTROLLED]
        Tab 1: <Tab_Title> (<URL>)
        Tab 2: <Tab_Title> (<URL>)

        Returns:
            Tuple containing:
            - int: The number of tabs
            - str: String describing each tab.
        """
        num_tabs = 1
        assert self._context is not None
        assert self._page is not None
        tabs_information = await self._playwright_controller.get_tabs_information(
            self._context,
            self._page,  # Pass the current page
        )
        num_tabs = len(tabs_information)
        tabs_information_str = "\n".join(
            [
                f"Tab {tab['index']}: {tab['title']} ({tab['url']})"
                f"{' [CURRENTLY SHOWN]' if tab['is_active'] else ''}"
                f"{' [CONTROLLED]' if tab['is_controlled'] else ''}"
                for tab in tabs_information
            ]
        )
        return num_tabs, tabs_information_str

    def _to_config(self) -> WebSurferConfig:
        return WebSurferConfig(
            name=self.name,
            model_client=self._model_client.dump_component(),
            browser=self._browser.dump_component(),
            downloads_folder=self.downloads_folder,
            description=self.description,
            debug_dir=self.debug_dir,
            start_page=self.start_page,
            animate_actions=self.animate_actions,
            to_save_screenshots=self.to_save_screenshots,
            max_actions_per_step=self.max_actions_per_step,
            to_resize_viewport=self.to_resize_viewport,
            url_statuses=self._url_status_manager.url_statuses,
            single_tab_mode=self.single_tab_mode,
            json_model_output=self.json_model_output,
            multiple_tools_per_call=self.multiple_tools_per_call,
            viewport_height=self.viewport_height,
            viewport_width=self.viewport_width,
            use_action_guard=self.use_action_guard,
        )

    @classmethod
    def _from_config(cls, config: WebSurferConfig) -> Self:
        return cls(
            name=config.name,
            model_client=ChatCompletionClient.load_component(config.model_client),
            browser=PlaywrightBrowser.load_component(config.browser),
            model_context_token_limit=config.model_context_token_limit,
            downloads_folder=config.downloads_folder,
            description=config.description or cls.DEFAULT_DESCRIPTION,
            debug_dir=config.debug_dir,
            start_page=config.start_page or cls.DEFAULT_START_PAGE,
            animate_actions=config.animate_actions,
            to_save_screenshots=config.to_save_screenshots,
            to_resize_viewport=config.to_resize_viewport,
            url_statuses=config.url_statuses,
            url_block_list=config.url_block_list,
            single_tab_mode=config.single_tab_mode,
            json_model_output=config.json_model_output,
            multiple_tools_per_call=config.multiple_tools_per_call,
            viewport_height=config.viewport_height,
            viewport_width=config.viewport_width,
            use_action_guard=config.use_action_guard,
        )

    @classmethod
    def from_config(cls, config: WebSurferConfig) -> Self:
        return cls._from_config(config)

    async def save_state(self) -> Mapping[str, Any]:
        """Save the current state of the WebSurfer.

        Returns:
            A dictionary containing the chat history and browser state
        """
        assert self._context is not None
        # Get the browser state and convert it to a dict
        browser_state = await save_browser_state(self._context, self._page)

        # Create and return the WebSurfer state
        state = WebSurferState(
            chat_history=self._chat_history,
            browser_state=browser_state,
        )
        return state.model_dump()

    async def load_state(self, state: Mapping[str, Any]) -> None:
        """Load a previously saved state.

        Args:
            state: Dictionary containing the state to load
        """
        # Validate and convert the state to a WebSurferState
        web_surfer_state = WebSurferState.model_validate(state)

        # Update the chat history
        self._chat_history = web_surfer_state.chat_history

        # Load the browser state if it exists
        if web_surfer_state.browser_state is not None:
            assert self._context is not None
            await load_browser_state(self._context, web_surfer_state.browser_state)

            # Update the controlled page to the active tab
            if self._context and self._context.pages:
                active_index = min(
                    web_surfer_state.browser_state.activeTabIndex,
                    len(self._context.pages) - 1,
                )
                self._page = self._context.pages[active_index]



# /api/plans routes
from fastapi import APIRouter, Depends, HTTPException
from loguru import logger
import os
import yaml
from pathlib import Path
from typing import Dict, Any
from pydantic import BaseModel

from autogen_agentchat.messages import TextMessage, MultiModalMessage
from autogen_core.models import ChatCompletionClient

from ....learning import learn_plan_from_messages
from ....learning.memory_provider import MemoryControllerProvider

from ...datamodel import Plan
from ..deps import get_db
from .sessions import list_session_runs

router = APIRouter()


@router.get("/")
async def list_plans(user_id: str, db=Depends(get_db)) -> Dict:
    """Get all plans for a user"""
    response = db.get(Plan, filters={"user_id": user_id})
    return {"status": True, "data": response.data}


@router.get("/{plan_id}")
async def get_plan(plan_id: int, user_id: str, db=Depends(get_db)) -> Dict:
    """Get a specific plan"""
    response = db.get(Plan, filters={"id": plan_id, "user_id": user_id})
    if not response.status or not response.data:
        raise HTTPException(status_code=404, detail="Plan not found")
    return {"status": True, "data": response.data[0]}


@router.post("/")
async def create_plan(plan: Plan, db=Depends(get_db)) -> Dict:
    """Create a new plan"""
    if not plan.user_id:
        raise HTTPException(status_code=400, detail="user_id is required")

    plan_response = db.upsert(plan)
    if not plan_response.status:
        raise HTTPException(status_code=400, detail=plan_response.message)

    return {"status": True, "data": plan_response.data}


@router.put("/{plan_id}")
async def update_plan(
    plan_id: int, user_id: str, plan: Plan, db=Depends(get_db)
) -> Dict:
    existing_plan = db.get(Plan, filters={"id": plan_id, "user_id": user_id})
    if not existing_plan.status or not existing_plan.data:
        raise HTTPException(status_code=404, detail="Plan not found")

    response = db.upsert(plan)
    if not response.status:
        raise HTTPException(status_code=400, detail=response.message)
    return {
        "status": True,
        "data": response.data,
        "message": "Plan updated successfully",
    }


@router.delete("/{plan_id}")
async def delete_plan(plan_id: int, user_id: str, db=Depends(get_db)) -> Dict:
    """Delete a specific plan"""
    response = db.delete(Plan, filters={"id": plan_id, "user_id": user_id})
    if not response.status:
        raise HTTPException(status_code=400, detail=response.message)
    return {"status": True, "data": response.data}


# Create a request model
class LearnPlanRequest(BaseModel):
    session_id: int
    user_id: str


@router.post("/learn_plan")
async def learn_plan(
    request: LearnPlanRequest,
    db=Depends(get_db),
):
    """Learn a plan from chat messages in a session"""
    session_id = request.session_id
    user_id = request.user_id

    if not session_id or not user_id:
        return {
            "status": False,
            "message": "Missing required parameters",
            "error": "MISSING_PARAMETERS",
        }

    try:
        # Read the config file, if present
        config: dict[str, Any] = {}
        config_file = os.environ.get("_CONFIG")
        if config_file:
            with open(config_file, "r") as f:
                config = yaml.safe_load(f)

        # Load the client from the config
        plan_learning_config = config.get("plan_learning_client", None)
        if not plan_learning_config:
            # Fallback to orchestrator_client if plan_learning_client is not set
            plan_learning_config = config.get("orchestrator_client", None)

        if plan_learning_config:
            model_client = ChatCompletionClient.load_component(plan_learning_config)
        else:
            # If nothing was provided, use a safe default
            model_client = ChatCompletionClient.load_component(
                {
                    "provider": "drsai.HepAIChatCompletionClient",
                    "config": {
                        "model": "openai/gpt-4o",
                    },
                    "max_retries": 5,
                }
            )

        # 1. Retrieve messages from database
        runs_result = await list_session_runs(
            session_id=session_id, user_id=user_id, db=db
        )

        runs = runs_result.get("data", {}).get("runs", [])
        messages = []
        if len(runs) > 0:
            messages = runs[0].get("messages", [])  # only 1 run per session

        if not messages:
            return {
                "status": False,
                "message": "No messages found in this session",
                "error": "NO_MESSAGES",
            }

        # 2. Format messages for learn_plan
        messages_for_learning = []
        for msg in messages:
            # Skip messages from non-agent or orchestrator sources
            if msg.config.get("source", "") not in [
                "user",
                "user_proxy",
                "web_surfer",
                "file_surfer",
                "orchestrator",
                "Orchestrator",
                "coder_agent-llm",
                "coder_agent-executor",
            ]:
                continue
            if msg.config.get("type") == "TextMessage":
                messages_for_learning.append(
                    TextMessage(
                        source=msg.config.get("source", ""),
                        content=msg.config.get("content", ""),
                    )
                )
            elif msg.config.get("type") == "MultiModalMessage":
                messages_for_learning.append(
                    MultiModalMessage(
                        source=msg.config.get("source", ""),
                        content=msg.config.get("content", []),
                    )
                )

        # 3. Call learn_plan

        plan = await learn_plan_from_messages(model_client, messages_for_learning)

        # 4. Convert PlanStep objects to dictionaries
        steps_as_dicts = []
        for step in plan.steps:
            if isinstance(step, dict):
                steps_as_dicts.append(step)
            else:
                try:
                    steps_as_dicts.append(step.model_dump())
                except AttributeError:
                    step_dict = {
                        "title": getattr(step, "title", ""),
                        "details": getattr(step, "details", ""),
                        "agent_name": getattr(step, "agent_name", ""),
                    }
                    steps_as_dicts.append(step_dict)

        # Create database plan with converted steps
        db_plan = Plan(
            task=plan.task, steps=steps_as_dicts, user_id=user_id, session_id=session_id
        )
        response = db.upsert(db_plan)

        # Add the plan to memory
        try:
            memory_provider = MemoryControllerProvider(
                internal_workspace_root=Path(os.environ.get("INTERNAL_WORKSPACE_ROOT")),
                external_workspace_root=Path(os.environ.get("EXTERNAL_WORKSPACE_ROOT")),
                inside_docker=os.environ.get("INSIDE_DOCKER", "false").lower()
                == "true",
            )
            memory_controller = memory_provider.get_memory_controller(
                user_id, model_client
            )

            logger.info("Adding plan to memory...")
            await memory_controller.add_memo(
                task=plan.task, insight=plan.model_dump_json(), index_on_both=False
            )
            logger.info("Plan successfully added to memory")

        except Exception as e:
            logger.error(f"Error adding plan to memory: {e}")

        return {
            "status": True,
            "data": {
                "planId": response.data.get("id") if response.data else None,
            },
            "message": "Plan created successfully",
        }

    except Exception:
        import traceback

        logger.error(traceback.format_exc())

        return {
            "status": False,
            "message": "Failed to create plan",
            "error": "INTERNAL_ERROR",
        }



from __future__ import annotations

import asyncio
import json
import logging
import os
import time
from pathlib import Path
from typing import (
    AsyncGenerator,
    Any,
    cast,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
)
import aiofiles
import yaml
from loguru import logger
from autogen_agentchat.base import ChatAgent, TaskResult, Team
from autogen_agentchat.messages import AgentEvent, ChatMessage, TextMessage
from autogen_core import EVENT_LOGGER_NAME, CancellationToken, ComponentModel
from autogen_core.logging import LLMCallEvent
from ...types import RunPaths

from ...input_func import InputFuncType

from ..datamodel.types import EnvironmentVariable, LLMCallEventMessage, TeamResult
from ..datamodel.db import Run
from ..utils.utils import get_modified_files

from ....agent_factory.magentic_one.task_team import create_magentic_one_team, create_magentic_round_team

from dotenv import load_dotenv
load_dotenv()


class RunEventLogger(logging.Handler):
    """Event logger that queues LLMCallEvents for streaming"""

    def __init__(self) -> None:
        super().__init__()
        self.events: asyncio.Queue[LLMCallEventMessage] = asyncio.Queue()

    def emit(self, record: logging.LogRecord) -> None:
        if isinstance(record.msg, LLMCallEvent):
            self.events.put_nowait(LLMCallEventMessage(content=str(record.msg)))


class TeamManager:
    """Manages team operations including loading configs and running teams"""

    def __init__(
        self,
        internal_workspace_root: Path,
        external_workspace_root: Path,
        inside_docker: bool = True,
        config: dict[str, Any] = {},
    ) -> None:
        self.team: Team | ChatAgent | None = None
        self.load_from_config = False
        self.internal_workspace_root = internal_workspace_root
        self.external_workspace_root = external_workspace_root
        self.inside_docker = inside_docker

        self.config = config
        self.novnc_port: int|None = None
        self.playwright_port: int|None = None
        self.mode: str|None = None

    @staticmethod
    async def load_from_file(path: Union[str, Path]) -> Dict[str, Any]:
        """Load team configuration from JSON/YAML file"""
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Config file not found: {path}")

        async with aiofiles.open(path) as f:
            content = await f.read()
            if path.suffix == ".json":
                return json.loads(content)
            elif path.suffix in (".yml", ".yaml"):
                return yaml.safe_load(content)
            raise ValueError(f"Unsupported file format: {path.suffix}")

    def prepare_run_paths(
        self,
        run: Optional[Run] = None,
    ) -> RunPaths:
        external_workspace_root = self.external_workspace_root
        internal_workspace_root = self.internal_workspace_root

        if run:
            run_suffix = os.path.join(
                "files",
                "user",
                str(run.user_id or "unknown_user"),
                str(run.session_id or "unknown_session"),
                str(run.id or "unknown_run"),
            )
        else:
            run_suffix = os.path.join(
                "files", "user", "unknown_user", "unknown_session", "unknown_run"
            )

        internal_run_dir = internal_workspace_root / Path(run_suffix)
        external_run_dir = external_workspace_root / Path(run_suffix)
        # Can only make dir on internal, as it is what a potential docker container sees.
        # TO-ANSWER: why?
        logger.info(f"Creating run dirs: {internal_run_dir} and {external_run_dir}")
        if self.inside_docker:
            internal_run_dir.mkdir(parents=True, exist_ok=True)
        else:
            external_run_dir.mkdir(parents=True, exist_ok=True)

        return RunPaths(
            internal_root_dir=internal_workspace_root,
            external_root_dir=external_workspace_root,
            run_suffix=run_suffix,
            internal_run_dir=internal_run_dir,
            external_run_dir=external_run_dir,
        )

    @staticmethod
    async def load_from_directory(directory: Union[str, Path]) -> List[Dict[str, Any]]:
        """Load all team configurations from a directory"""
        directory = Path(directory)
        configs: List[Dict[str, Any]] = []
        valid_extensions = {".json", ".yaml", ".yml"}

        for path in directory.iterdir():
            if path.is_file() and path.suffix.lower() in valid_extensions:
                try:
                    config = await TeamManager.load_from_file(path)
                    configs.append(config)
                except Exception as e:
                    logger.error(f"Failed to load {path}: {e}")

        return configs

    async def _create_team(
        self,
        team_config: Union[str, Path, Dict[str, Any], ComponentModel],
        state: Optional[Mapping[str, Any] | str] = None,
        input_func: Optional[InputFuncType] = None,
        env_vars: Optional[List[EnvironmentVariable]] = None,
        settings_config: dict[str, Any] = {},
        *,
        paths: RunPaths,
        run: Optional[Run] = None,
        files: List[Dict[str, Any]] | None = None,
    ) -> None:
        """Create team instance from config"""

        # TODO: 判断前端传递的settings_config中agent的配置, 在src/drsai/agent_factory/load_agent.py中进行系统性判断和配置的加载

        try:
            if not self.load_from_config:
                # Load model configurations from settings if provided
                model_configs: Dict[str, Any] = {}
                if isinstance(settings_config.get("model_configs"), str):
                    try:
                        model_configs = yaml.safe_load(settings_config["model_configs"])
                    except Exception as e:
                        logger.error(f"Error loading model configs: {e}")
                        raise e
                settings_config["model_configs"] = model_configs

                # 判断前端的agent mode配置, 并进行相应的agent/team的创建
                agent_mode_config: dict[str, Any] = settings_config.pop("agent_mode_config", {})
                self.mode = agent_mode_config.get("mode", "magentic-one")
                
                if self.mode in ["magentic-one"]:
                    settings_config["model_configs"] = model_configs
                    team, novnc_port, playwright_port = await create_magentic_one_team(
                        team_config = team_config,
                        state = state,
                        input_func = input_func,
                        env_vars = env_vars,
                        settings_config = settings_config,
                        paths = paths,
                        config = self.config,
                        load_from_config = self.load_from_config,
                        inside_docker = self.inside_docker,
                    )
                    self.novnc_port = novnc_port
                    self.playwright_port = playwright_port
                else:
                    team, novnc_port, playwright_port = await create_magentic_round_team(
                        team_config = team_config,
                        state = state,
                        input_func = input_func,
                        env_vars = env_vars,
                        settings_config = settings_config,
                        paths = paths,
                        config = self.config,
                        load_from_config = self.load_from_config,
                        inside_docker = self.inside_docker,
                        run_info = run.model_dump() if run else {},
                        agent_mode_config = agent_mode_config,
                        files = files
                    )
                    self.novnc_port = novnc_port
                    self.playwright_port = playwright_port
            else:
                raise NotImplementedError("load_from_config not implemented yet")
            self.team = team
        except Exception as e:
            logger.error(f"Error creating team: {e}")
            await self.close()
            raise
        
    # async def _create_team(
    #     self,
    #     team_config: Union[str, Path, Dict[str, Any], ComponentModel],
    #     state: Optional[Mapping[str, Any] | str] = None,
    #     input_func: Optional[InputFuncType] = None,
    #     env_vars: Optional[List[EnvironmentVariable]] = None,
    #     settings_config: dict[str, Any] = {},
    #     *,
    #     paths: RunPaths,
    # ) -> tuple[Team, int, int]:
    #     """Create team instance from config"""

    #     _, novnc_port, playwright_port = get_browser_resource_config(
    #         paths.external_run_dir, -1, -1, self.inside_docker
    #     )
    #     try:
    #         if not self.load_from_config:
    #             # Load model configurations from settings if provided
    #             model_configs: Dict[str, Any] = {}
    #             if settings_config.get("model_configs"):
    #                 try:
    #                     model_configs = yaml.safe_load(settings_config["model_configs"])
    #                 except Exception as e:
    #                     logger.error(f"Error loading model configs: {e}")
    #                     raise e

    #             # Use model configs from settings if available, otherwise fall back to config
    #             orchestrator_config = model_configs.get("orchestrator_client", self.config.get("orchestrator_client", None))
    #             model_client_configs = ModelClientConfigs(
    #                 orchestrator=orchestrator_config,
    #                 web_surfer=model_configs.get(
    #                     "web_surfer_client",
    #                     self.config.get("web_surfer_client", None),
    #                 ),
    #                 coder=model_configs.get(
    #                     "coder_client", self.config.get("coder_client", None)
    #                 ),
    #                 file_surfer=model_configs.get(
    #                     "file_surfer_client",
    #                     self.config.get("file_surfer_client", None),
    #                 ),
    #                 action_guard=model_configs.get(
    #                     "action_guard_client",
    #                     self.config.get("action_guard_client", None),
    #                 ),
    #             )

    #             magentic_ui_config = MagenticUIConfig(
    #                 **(settings_config or {}),
    #                 model_client_configs=model_client_configs,
    #                 playwright_port=playwright_port,
    #                 novnc_port=novnc_port,
    #                 inside_docker=self.inside_docker,
    #             )

    #             self.team = cast(
    #                 Team,
    #                 await get_task_team(
    #                     magentic_ui_config=magentic_ui_config,
    #                     input_func=input_func,
    #                     paths=paths,
    #                 ),
    #             )
    #             if hasattr(self.team, "_participants"):
    #                 for agent in cast(list[ChatAgent], self.team._participants):  # type: ignore
    #                     if isinstance(agent, WebSurfer):
    #                         novnc_port = agent.novnc_port
    #                         playwright_port = agent.playwright_port

    #             if state:
    #                 if isinstance(state, str):
    #                     try:
    #                         # Try to decompress if it's compressed
    #                         state_dict = decompress_state(state)
    #                         await self.team.load_state(state_dict)
    #                     except Exception:
    #                         # If decompression fails, assume it's a regular JSON string
    #                         state_dict = json.loads(state)
    #                         await self.team.load_state(state_dict)
    #                 else:
    #                     await self.team.load_state(state)

    #             return self.team, novnc_port, playwright_port

    #         if isinstance(team_config, (str, Path)):
    #             config = await self.load_from_file(team_config)
    #         elif isinstance(team_config, dict):
    #             config = team_config
    #         else:
    #             config = team_config.model_dump()

    #         # Load env vars into environment if provided
    #         if env_vars:
    #             logger.info("Loading environment variables")
    #             for var in env_vars:
    #                 os.environ[var.name] = var.value

    #         self.team = cast(Team, GroupChat.load_component(config))

    #         if hasattr(self.team, "_participants"):
    #             for agent in cast(list[ChatAgent], self.team._participants):  # type: ignore
    #                 if hasattr(agent, "input_func"):
    #                     agent.input_func = input_func  # type: ignore
    #                 if isinstance(agent, WebSurfer):
    #                     novnc_port = agent.novnc_port or -1
    #                     playwright_port = agent.playwright_port or -1
    #         return self.team, novnc_port, playwright_port
    #     except Exception as e:
    #         logger.error(f"Error creating team: {e}")
    #         await self.close()
    #         raise

    async def run_stream(
        self,
        task: Optional[Union[ChatMessage, str, Sequence[ChatMessage]]],
        team_config: Union[str, Path, dict[str, Any], ComponentModel],
        state: Optional[Mapping[str, Any] | str] = None,
        input_func: Optional[InputFuncType] = None,
        cancellation_token: Optional[CancellationToken] = None,
        env_vars: Optional[List[EnvironmentVariable]] = None,
        settings_config: Optional[Dict[str, Any]] = None,
        run: Optional[Run] = None,
        files: List[Dict[str, Any]] | None = None,
    ) -> AsyncGenerator[
        Union[AgentEvent, ChatMessage, LLMCallEventMessage, TeamResult], None
    ]:
        """Stream team execution results"""
        start_time = time.time()

        # Setup logger correctly
        logger = logging.getLogger(EVENT_LOGGER_NAME)
        logger.setLevel(logging.CRITICAL)
        llm_event_logger = RunEventLogger()
        logger.handlers = [llm_event_logger]  # Replace all handlers
        logger.info(f"Running in docker: {self.inside_docker}")
        paths = self.prepare_run_paths(run=run)
        known_files = set(
            file["name"]
            for file in get_modified_files(
                0, time.time(), source_dir=str(paths.internal_run_dir)
            )
        )
        global_new_files: List[Dict[str, str]] = []
        try:
            # TODO: This might cause problems later if we are not careful
        
            # mode = settings_config.get("Use_default_Agent_Groupchat_mode", None)

            if self.team is None:
                # TODO: if we start allowing load from config, we'll need to write the novnc and playwright ports back to the team config..
                await self._create_team(
                    team_config,
                    state,
                    input_func,
                    env_vars,
                    settings_config or {},
                    paths=paths,
                    run=run,
                    files=files,
                )

                # Initialize known files by name for tracking
                initial_files = get_modified_files(
                    start_time, time.time(), source_dir=str(paths.internal_run_dir)
                )
                known_files = {file["name"] for file in initial_files}

               
                if self.mode in ["magentic-one"]:
                    # 前端启动noVNC，在drsai模式下不需要启动noVNC
                    VNC_SERVICE_URL = os.getenv("VNC_SERVICE_URL", "localhost")
                    yield TextMessage(
                        source="system",
                        content=f"Browser noVNC address can be found at {VNC_SERVICE_URL}:{self.novnc_port}/vnc.html",
                        metadata={
                            "internal": "no",
                            "type": "browser_address",
                            "novnc_port": str(self.novnc_port),
                            "playwright_port": str(self.playwright_port),
                        },
                    )

                async for message in self.team.run_stream(  # type: ignore
                    task=task, cancellation_token=cancellation_token
                ):
                    if cancellation_token and cancellation_token.is_cancelled():
                        break

                    # Get all current files with full metadata
                    modified_files = get_modified_files(
                        start_time, time.time(), source_dir=str(paths.internal_run_dir)
                    )
                    current_file_names = {file["name"] for file in modified_files}

                    # Find new files
                    new_file_names = current_file_names - known_files
                    known_files = current_file_names  # Update for next iteration

                    # Get the full data for new files
                    new_files = [
                        file
                        for file in modified_files
                        if file["name"] in new_file_names
                    ]

                    if new_files:
                        # filter files that start with "tmp_code"
                        new_files = [
                            file
                            for file in new_files
                            if not file["name"].startswith("tmp_code")
                        ]
                        if len(new_files) > 0:
                            file_message = TextMessage(
                                source="system",
                                content="File Generated",
                                metadata={
                                    "internal": "no",
                                    "type": "file",
                                    "files": json.dumps(new_files),
                                },
                            )
                            global_new_files.extend(new_files)
                            yield file_message

                    if isinstance(message, TaskResult):
                        yield TeamResult(
                            task_result=message,
                            usage="",
                            duration=time.time() - start_time,
                            files=modified_files,  # Full file data preserved
                        )
                    else:
                        yield message

                    # Add generated files to final output
                    if (
                        isinstance(message, TextMessage)
                        and message.metadata.get("type", "") == "final_answer"
                    ):
                        if len(global_new_files) > 0:
                            # only keep unique file names, if there is a file with the same name, keep the latest one
                            global_new_files = list(
                                {
                                    file["name"]: file for file in global_new_files
                                }.values()
                            )
                            file_message = TextMessage(
                                source="system",
                                content="File Generated",
                                metadata={
                                    "internal": "no",
                                    "type": "file",
                                    "files": json.dumps(global_new_files),
                                },
                            )
                            yield file_message
                            global_new_files = []

                    # Check for any LLM events
                    while not llm_event_logger.events.empty():
                        event = await llm_event_logger.events.get()
                        yield event
        finally:
            # Cleanup - remove our handler
            if llm_event_logger in logger.handlers:
                logger.handlers.remove(llm_event_logger)

            # Ensure cleanup happens
            if self.team and hasattr(self.team, "close"):
                logger.info("Closing team")
                await self.team.close()  # type: ignore
                logger.info("Team closed")

    async def close(self):
        """Close the team manager"""
        if self.team and hasattr(self.team, "close"):
            logger.info("Closing team")
            await self.team.close()  # type: ignore
            self.team = None
            logger.info("Team closed")
        else:
            logger.warning("Team manager is not initialized or already closed")

    async def run(
        self,
        task: ChatMessage | Sequence[ChatMessage] | str | None,
        team_config: Union[str, Path, dict[str, Any], ComponentModel],
        input_func: Optional[InputFuncType] = None,
        cancellation_token: Optional[CancellationToken] = None,
        env_vars: Optional[List[EnvironmentVariable]] = None,
    ) -> TeamResult:
        """Run team synchronously"""
        raise NotImplementedError("Use run_stream instead")

    async def pause_run(self) -> None:
        """Pause the run"""
        if self.team:
            await self.team.pause()

    async def resume_run(self) -> None:
        """Resume the run"""
        if self.team:
            await self.team.resume()
